// List of sets -*- c++ -*-

#ifndef SETLIST_H_
# define SETLIST_H_
# ifdef __GNUC__
#  pragma interface
# endif // __GNUC__

# include <cassert>

/** @file SetList.h
 * A collection of sets of numbers
 */

/** List of sets */
class SetList
{
public:
  /** Constructor
   * @param numSets	number of sets
   */
  explicit SetList (unsigned numSets);
private:
  /** Copy constructor */
  SetList (const class SetList& old);
  /** Assignment operator */
  class SetList& operator= (const class SetList& old);
public:
  /** Destructor */
  ~SetList ();
  /** Assign an array to a set
   * @param index	index number of the set
   * @param array	array to be assigned (length==array[0])
   */
  void assign (unsigned index, unsigned* array);
  /** Assign a bit vector to a set
   * @param index	index number to the set
   * @param bv		BitVector to be assigned
   * @param offset	offset for index numbers
   * @param complement	flag: treat BitVector as its complement
   */
  void assign (unsigned index,
	       const class BitVector &bv,
	       unsigned offset,
	       bool complement);
  /** Copy an array to a set
   * @param index	index number of the set
   * @param array	array to be copied (length==array[0])
   */
  void copy (unsigned index, const unsigned* array);
  /** Grow the list
   * @param numSets	new size of the list
   */
  void grow (unsigned numSets);
  /** Shrink the list
   * @param numSets	new size of the list
   */
  void shrink (unsigned numSets);

  /** Determine the number of sets */
  unsigned getNumSets () const { return myNumSets; }
  /** Determine the size of a set
   * @param index	index number of the set
   * @return		number of elements in the set
   */
  unsigned getSize (unsigned index) const {
    assert (index < myNumSets);
    return mySets[index] ? *mySets[index] : 0;
  }
  /** Access a set
   * @param index	index number of the set
   * @return		pointer to the first element in the set
   */
  const unsigned* operator[] (unsigned index) const {
    assert (index < myNumSets);
    return mySets[index] + 1;
  }
  /** Access a set
   * @param index	index number of the set
   * @return		pointer to the first element in the set
   */
  unsigned* operator[] (unsigned index) {
    assert (index < myNumSets);
    return mySets[index] + 1;
  }
  /** Remove an element of a set by assigning the last element to its place
   * @param index	index number of the set
   * @param elem	index of the element
   */
  void delElement (unsigned index, unsigned elem) {
    assert(index < myNumSets && elem < *mySets[index]);
    mySets[index][elem + 1] = mySets[index][(*mySets[index])--];
  }

private:
  /** Number of sets */
  unsigned myNumSets;
  /** Number of allocated sets */
  unsigned myAllocatedSets;
  /** The sets */
  unsigned** mySets;
};

#endif // SETLIST_H_
