// B-tree map of numbers to numbers -*- c++ -*-

#ifdef __GNUC__
# pragma implementation
#endif // __GNUC__
#include "BTree.h"
#include <cassert>

/** @file BTree.C
 * B-tree map of numbers to numbers
 */

/** Size of a b-tree page, in bytes */
#define PAGE_SIZE (sizeof (item_t) * BTREE_SIZE)

/** Flag for leaf items */
#define PAGE_LEAF ~((~0u) >> 1)

/** Determine the amount of keys in a B-tree node */
inline static unsigned
keys (const BTree::item_t item) { return item & ~PAGE_LEAF; }
/** Determine whether a B-tree node is a leaf node */
inline static bool
isLeaf (const BTree::item_t& item) { return item & PAGE_LEAF; }

BTree::BTree (FILE* file) :
  myFile (file)
{
  if (1 != fread (myRoot, PAGE_SIZE, 1, myFile)) {
    memset (myRoot, 0, PAGE_SIZE);
    myRoot[0] = PAGE_LEAF;
    fseek (myFile, 0, SEEK_SET);
    fwrite (myRoot, PAGE_SIZE, 1, myFile);
  }
}

BTree::~BTree ()
{
  fclose (myFile);
}

void
BTree::insert (item_t key, item_t value)
{
  if (keys (*myRoot) == BTREE_SIZE / 2 - 1) {
    item_t page[BTREE_SIZE];
    memcpy (page, myRoot, PAGE_SIZE);
    memset (myRoot, 0, PAGE_SIZE);
    myRoot[0] = 0;
    fseek (myFile, 0, SEEK_END);
    split (myRoot, page, 1, 0,
	   myRoot[BTREE_SIZE / 2] = ftell (myFile) / PAGE_SIZE);
  }
  insert (0, key, value);
}

BTree::item_t*
BTree::search (const item_t* page, item_t key) const
{
  item_t p[BTREE_SIZE], p2[BTREE_SIZE];
  for (;;) {
    unsigned index;
    for (index = 1; index <= keys (*page); index++) {
      if (page[index] < key)
	continue;
      if (page[index] != key)
	break;
      item_t* result;
      unsigned index2 = index;
      while (++index2 <= keys (*page) && page[index2] == key);
      if (isLeaf (*page)) {
	index2 -= index;
	*(result = new item_t[1 + index2]) = index2;
	memcpy (result + 1, &page[BTREE_SIZE / 2 + index],
		index2 * sizeof *result);
	return result;
      }
      else {
	for (result = 0; index2 >= index; index2--) {
	  fseek (myFile, page[(BTREE_SIZE / 2 - 1) + index2] *
		 PAGE_SIZE, SEEK_SET);
	  fread (p2, PAGE_SIZE, 1, myFile);
	  if (item_t* r = search (p2, key)) {
	    if (!result) result = r;
	    else {
	      item_t* r2 = new item_t[*result + *r + 1];
	      *r2 = *result + *r;
	      memcpy (r2 + 1, result + 1, *result * sizeof *result);
	      memcpy (r2 + 1 + *result, r + 1, *r * sizeof *r);
	      delete[] result;
	      delete[] r;
	      result = r2;
	    }
	  }
	}
	return result;
      }
    }
    if (isLeaf (*page))
      return 0;
    assert (page[(BTREE_SIZE / 2 - 1) + index] > 0);
    fseek (myFile, page[(BTREE_SIZE / 2 - 1) + index] * PAGE_SIZE, SEEK_SET);
    fread (p, PAGE_SIZE, 1, myFile);
    page = p;
  }
}

BTree::item_t*
BTree::search (item_t* page, item_t key) const
{
  for (;;) {
    unsigned index;
    for (index = 1; index <= keys (*page); index++) {
      if (page[index] < key)
	continue;
      if (page[index] != key)
	break;
      item_t* result;
      unsigned index2 = index;
      while (++index2 <= keys (*page) && page[index2] == key);
      if (isLeaf (*page)) {
	index2 -= index;
	*(result = new item_t[1 + index2]) = index2;
	memcpy (result + 1, &page[BTREE_SIZE / 2 + index],
		index2 * sizeof *result);
	return result;
      }
      else {
	item_t p[BTREE_SIZE];
	for (result = 0; index2 >= index; index2--) {
	  fseek (myFile, page[(BTREE_SIZE / 2 - 1) + index2] * PAGE_SIZE,
		 SEEK_SET);
	  fread (p, PAGE_SIZE, 1, myFile);
	  if (item_t* r = search (p, key)) {
	    if (!result) result = r;
	    else {
	      item_t* r2 = new item_t[*result + *r + 1];
	      *r2 = *result + *r;
	      memcpy (r2 + 1, result + 1, *result * sizeof *result);
	      memcpy (r2 + 1 + *result, r + 1, *r * sizeof *r);
	      delete[] result;
	      delete[] r;
	      result = r2;
	    }
	  }
	}
	return result;
      }
    }
    if (isLeaf (*page))
      return 0;
    assert (page[(BTREE_SIZE / 2 - 1) + index] > 0);
    fseek (myFile, page[(BTREE_SIZE / 2 - 1) + index] * PAGE_SIZE, SEEK_SET);
    fread (page, PAGE_SIZE, 1, myFile);
  }
}

void
BTree::split (item_t* parent, item_t* child,
	      unsigned index,
	      unsigned parentpage,
	      unsigned childpage)
{
  assert (keys (*child) == BTREE_SIZE / 2 - 1 &&
	  index && index - 1 <= keys (*parent));
  item_t child2[BTREE_SIZE];
  memset (child2, 0, PAGE_SIZE);
  memcpy (child2 + 1, child + (BTREE_SIZE / 4 + 1),
	  (BTREE_SIZE / 4 - 1) * sizeof *child);
  memcpy (child2 + BTREE_SIZE / 2, child + (BTREE_SIZE * 3 / 4),
	  (BTREE_SIZE / 4) * sizeof *child);
  if (isLeaf (*child)) {
    child[0] = (BTREE_SIZE / 4) | PAGE_LEAF;
    child2[0] = (BTREE_SIZE / 4 - 1) | PAGE_LEAF;
  }
  else
    child2[0] = child[0] = BTREE_SIZE / 4 - 1;
  if (index <= keys (*parent)) {
    memmove (parent + index + (BTREE_SIZE / 2 + 1),
	     parent + index + BTREE_SIZE / 2,
	     (keys (*parent) + 1 - index) * sizeof *parent);
    memmove (parent + index + 1, parent + index,
	     (keys (*parent) + 1 - index) * sizeof *parent);
  }
  parent[index] = child[BTREE_SIZE / 4];
  parent[0]++;
  memset (child + (BTREE_SIZE / 4 + 1), 0,
	  (BTREE_SIZE / 4 - 1) * sizeof *child);
  if (isLeaf (*child))
    memset (child + (BTREE_SIZE * 3 / 4 + 1), 0,
	    (BTREE_SIZE / 4 - 1) * sizeof *child);
  else
    memset (child + (BTREE_SIZE * 3 / 4), 0,
	    (BTREE_SIZE / 4) * sizeof *child);
  fseek (myFile, childpage * PAGE_SIZE, SEEK_SET);
  fwrite (child, PAGE_SIZE, 1, myFile);
  fseek (myFile, 0, SEEK_END);
  parent[BTREE_SIZE / 2 + index] = ftell (myFile) / PAGE_SIZE;
  fwrite (child2, PAGE_SIZE, 1, myFile);
  fseek (myFile, parentpage * PAGE_SIZE, SEEK_SET);
  fwrite (parent, PAGE_SIZE, 1, myFile);
  if (!parentpage)
    memcpy (myRoot, parent, PAGE_SIZE);
}

void
BTree::insert (unsigned pagenbr, item_t key, item_t value)
{
  item_t page[BTREE_SIZE];
  for (;;) {
    if (pagenbr) {
      fseek (myFile, pagenbr * PAGE_SIZE, SEEK_SET);
      fread (page, PAGE_SIZE, 1, myFile);
    }
    else
      memcpy (page, myRoot, PAGE_SIZE);

    unsigned i = keys (*page);
    assert (i < BTREE_SIZE / 2 - 1);
    if (isLeaf (*page)) {
      while (i && key <= page[i]) {
	page[i + 1] = page[i];
	page[i + (BTREE_SIZE / 2 + 1)] = page[i + (BTREE_SIZE / 2)];
	i--;
      }
      page[i + 1] = key;
      page[i + (BTREE_SIZE / 2 + 1)] = value;
      page[0]++;
      fseek (myFile, pagenbr * PAGE_SIZE, SEEK_SET);
      fwrite (page, PAGE_SIZE, 1, myFile);
      if (!pagenbr)
	memcpy (myRoot, page, PAGE_SIZE);
      return;
    }

    while (i && key <= page[i]) i--;
    unsigned childpagenbr = page[(BTREE_SIZE / 2) + i++];
    assert (childpagenbr > 0);
    fseek (myFile, childpagenbr * PAGE_SIZE, SEEK_SET);
    item_t child[BTREE_SIZE];
    fread (child, PAGE_SIZE, 1, myFile);
    if (keys (*child) == (BTREE_SIZE / 2 - 1)) {
      split (page, child, i, pagenbr, childpagenbr);
      if (key > page[i]) i++;
      pagenbr = page[(BTREE_SIZE / 2 - 1) + i];
      assert (pagenbr > 0);
    }
    else
      pagenbr = childpagenbr;
  }
}
