// Net class -*- c++ -*-

#ifdef __GNUC__
# pragma implementation
#endif // __GNUC__
#include "Net.h"
#include "Place.h"
#include "Transition.h"
#include "Type.h"
#include "BoolType.h"
#include "IntType.h"
#include "CardType.h"
#include "CharType.h"
#include "GlobalMarking.h"
#include "Marking.h"
#include "Constant.h"
#include "Valuation.h"
#include "Graph.h"
#include "Search.h"
#include "Function.h"
#include "BooleanBinop.h"
#include "PlaceContents.h"
#include "LeafValue.h"
#include "Quantifier.h"
#include "Printer.h"
#include "HashGraph.h"
#include "StateList.h"

/** @file Net.C
 * Algebraic system net
 */

unsigned Net::theEdgeThresold = 0;

/** The predefined boolean type */
static class BoolType boolType;
/** The predefined signed integer type */
static class IntType intType;
/** The predefined unsigned integer type */
static class CardType cardType;
/** The predefined character type */
static class CharType charType;

/** Determine whether a type is predefined
 * @param type	type to be checked
 * @return	true if the type is predefined
 */
inline static bool
isPredefinedType (const class Type* type)
{
  return
    type == &boolType || type == &intType ||
    type == &cardType || type == &charType;
}

Net::Net () :
  myNumPlaces (0), myPlaces (0), myNumTransitions (0), myTransitions (0),
  myFunctionMap (),
  myPlaceMap (), myTransitionMap (), myTypeMap (), myTypeList (),
  myInitMarking (0),
  myReject (0), myDeadlock (0),
  myNumWeaklyFair (0), myNumStronglyFair (0), myNumMaxFair (0),
  myNumProps (0), myPropNames (0), myProps (0),
  myNumConstants (0), myConstantNames (0), myConstants (0)
{
}

Net::~Net ()
{
  while (myNumPlaces)
    delete myPlaces[--myNumPlaces];
  delete[] myPlaces;

  while (myNumTransitions)
    delete myTransitions[--myNumTransitions];
  delete[] myTransitions;

  for (FunctionMap::iterator f = myFunctionMap.begin ();
       f != myFunctionMap.end (); f++)
    delete f->second;

  for (TypeList::iterator t = myTypeList.begin ();
       t != myTypeList.end (); t++)
    if (!::isPredefinedType (*t))
      delete *t;

  delete myInitMarking;
  myReject->destroy ();
  myDeadlock->destroy ();
  while (myNumProps--) {
    delete[] myPropNames[myNumProps];
    myProps[myNumProps]->destroy ();
  }
  delete[] myPropNames;
  delete[] myProps;
  while (myNumConstants--) {
    delete[] myConstantNames[myNumConstants];
    myConstants[myNumConstants]->destroy ();
  }
  delete[] myConstantNames;
  delete[] myConstants;
}

void
Net::addFunction (class Function& function)
{
  assert (!getFunction (function.getName ()));
  myFunctionMap[function.getName ()] = &function;
}

bool
Net::addFairness (class Expression& qualifier,
		  bool isStrong)
{
  unsigned id = isStrong ? myNumStronglyFair : myNumWeaklyFair;
  if (qualifier.getKind () == Expression::eMarking) {
    class Valuation valuation;
    for (const class Marking* m =
	   static_cast<class Marking&>(qualifier).first ();
	 m; m = m->next ()) {
      if (m->getMultiplicity (valuation)) {
	for (unsigned tr = myNumTransitions; tr--; ) {
	  if (class Expression* expr =
	      m->getToken ()->disqualify (*myTransitions[tr])) {
	    unsigned n = myTransitions[tr]->addFairness (*expr, id, isStrong);
	    if (n > myNumMaxFair) myNumMaxFair = n;
	  }
	}
      }
      else if (!valuation.isOK ()) {
	qualifier.destroy ();
	return false;
      }
      id = isStrong ? ++myNumStronglyFair : ++myNumWeaklyFair;
    }
  }
  else {
    for (unsigned tr = myNumTransitions; tr--; ) {
      if (class Expression* expr = qualifier.disqualify (*myTransitions[tr])) {
	unsigned n = myTransitions[tr]->addFairness (*expr, id, isStrong);
	if (n > myNumMaxFair) myNumMaxFair = n;
      }
    }
    isStrong ? ++myNumStronglyFair : ++myNumWeaklyFair;
  }
  qualifier.destroy ();
  return true;
}

bool
Net::addFairness (class Expression& qualifier,
		  class Transition& transition,
		  bool isStrong)
{
  unsigned& id = isStrong ? myNumStronglyFair : myNumWeaklyFair;
  if (qualifier.getKind () == Expression::eMarking) {
    class Valuation valuation;
    for (const class Marking* m =
	   static_cast<class Marking&>(qualifier).first ();
	 m; m = m->next ()) {
      if (m->getMultiplicity (valuation)) {
	if (class Expression* expr =
	    m->getToken ()->disqualify (transition)) {
	  unsigned n = transition.addFairness (*expr, id++, isStrong);
	  if (n > myNumMaxFair) myNumMaxFair = n;
	}
      }
      else if (!valuation.isOK ()) {
	qualifier.destroy ();
	return false;
      }
    }
    qualifier.destroy ();
  }
  else {
    unsigned n = transition.addFairness (qualifier, id++, isStrong);
    if (n > myNumMaxFair) myNumMaxFair = n;
  }
  return true;
}

static bool
flagVisible (const class Expression& expr,
	     void* data)
{
  if (expr.getKind () == Expression::ePlaceContents)
    const_cast<class Place&>(static_cast<const class PlaceContents&>
			     (expr).getPlace ()).flagVisible ();
  return true;
}

void
Net::addReject (class Expression& expr)
{
  assert (!expr.isTemporal ());
  assert (expr.getKind () == Expression::eUndefined ||
	  expr.getType () && expr.getType ()->getKind () == Type::tBool);
  if (!expr.forVariables (&::flagVisible, 0))
    assert (false);
  if (myReject)
    myReject = BooleanBinop::construct (false, *myReject, expr);
  else
    myReject = &expr;
}

void
Net::addDeadlock (class Expression& expr)
{
  assert (!expr.isTemporal ());
  assert (expr.getKind () == Expression::eUndefined ||
	  expr.getType () && expr.getType ()->getKind () == Type::tBool);
  if (!expr.forVariables (&::flagVisible, 0))
    assert (false);
  if (myDeadlock)
    myDeadlock = BooleanBinop::construct (false, *myDeadlock, expr);
  else
    myDeadlock = &expr;
}

class Place&
Net::addPlace (char* name,
	       class Constraint* capacity,
	       const class Type& type)
{
  assert (!myInitMarking);
  assert (name && myPlaceMap.find (name) == myPlaceMap.end ());

  class Place* place = new class Place (myNumPlaces, name, capacity, type);

  myPlaceMap[place->getName ()] = place;

  if (!(myNumPlaces & (myNumPlaces + 1))) {
    class Place** p = new class Place*[(myNumPlaces + 1) << 1];
    memcpy (p, myPlaces, sizeof (*myPlaces) * myNumPlaces);
    delete[] myPlaces;
    myPlaces = p;
  }

  return *(myPlaces[myNumPlaces++] = place);
}

class Transition&
Net::addTransition (char* name)
{
  assert (!myInitMarking);
  assert (name && myTransitionMap.find (name) == myTransitionMap.end ());

  class Transition* transition = new class Transition (myNumTransitions, name);
  myTransitionMap[transition->getName ()] = transition;
  if (!(myNumTransitions & (myNumTransitions + 1))) {
    class Transition** t = new class Transition*[(myNumTransitions + 1) << 1];
    memcpy (t, myTransitions, sizeof (*myTransitions) * myNumTransitions);
    delete[] myTransitions;
    myTransitions = t;
  }
  return *(myTransitions[myNumTransitions++] = transition);
}

void
Net::addType (class Type& t, char* name)
{
  assert (!myInitMarking);
  class Type* type = &t;

  if (name) {
    if (type->getName ()) {
      assert (strcmp (type->getName (), name));
      type = type->copy ();
    }
    type->setName (name);
    assert (myTypeMap.find (type->getName ()) == myTypeMap.end ());
    myTypeMap[type->getName ()] = type;
  }

  if (!hasType (*type))
    myTypeList.push_back (type);
}

bool
Net::addProp (char* name, class Expression& prop)
{
  assert (name && !prop.isTemporal ());
  assert (prop.getType () && prop.getType ()->getKind () == Type::tBool);

  for (unsigned i = myNumProps; i--; )
    if (!strcmp (name, myPropNames[i]))
      return false;

  if (!(myNumProps & (myNumProps + 1))) {
    char** propNames = new char*[(myNumProps + 1) << 1];
    memcpy (propNames, myPropNames, myNumProps * sizeof *myPropNames);
    delete[] myPropNames;
    myPropNames = propNames;

    class Expression** props = new class Expression*[(myNumProps + 1) << 1];
    memcpy (props, myProps, myNumProps * sizeof *myProps);
    delete[] myProps;
    myProps = props;
  }

  myPropNames[myNumProps] = name;
  myProps[myNumProps++] = &prop;
  return true;
}

bool
Net::checkProps (const class GlobalMarking& m,
		 bool (&operation) (unsigned, const void*),
		 const void* data) const
{
  if (myNumProps) {
    class Valuation valuation;
    valuation.setGlobalMarking (&m);
    bool hold;

    for (unsigned i = 0; i < myNumProps; i++) {
      if (class Value* value = myProps[i]->eval (valuation)) {
	assert (value->getKind () == Value::vLeaf);
	hold = bool (static_cast<class LeafValue&>(*value));
	delete value;
	if (hold && !operation (i, data))
	  return false;
      }
      else
	valuation.clearErrors ();
    }
  }

  return true;
}

bool
Net::addConstant (char* name, class Constant& constant)
{
  assert (name && constant.getType () && !constant.getType ()->isLeaf ());

  for (unsigned i = myNumConstants; i--; )
    if (!strcmp (name, myConstantNames[i]))
      return false;

  if (!(myNumConstants & (myNumConstants + 1))) {
    char** constantNames = new char*[(myNumConstants + 1) << 1];
    memcpy (constantNames, myConstantNames,
	    myNumConstants * sizeof *myConstantNames);
    delete[] myConstantNames;
    myConstantNames = constantNames;

    class Constant** constants =
      new class Constant*[(myNumConstants + 1) << 1];
    memcpy (constants, myConstants, myNumConstants * sizeof *myConstants);
    delete[] myConstants;
    myConstants = constants;
  }

  myConstantNames[myNumConstants] = name;
  myConstants[myNumConstants++] = &constant;
  return true;
}

const class Constant*
Net::getConstant (const char* name) const
{
  for (unsigned i = myNumConstants; i--; )
    if (!strcmp (name, myConstantNames[i]))
      return myConstants[i];
  return 0;
}

/** Flag: has the analysis been interrupted? */
volatile bool interrupted = false;

/** Determine whether a state is rejected
 * @param m		the state
 * @param expr		the rejection condition
 * @param state		number of the state (for diagnostics)
 * @param printer	the printer object (for diagnostics)
 * @return		true if the state is fatally rejected
 */
bool
Net::isRejected (const class GlobalMarking& m,
		 card_t state,
		 const class Printer& printer) const
{
  if (!myReject) return false;
  class Valuation valuation;
  valuation.setGlobalMarking (&m);
  bool rejected = false;
  if (class Value* value = myReject->eval (valuation)) {
    assert (value->getKind () == Value::vLeaf);
    rejected = bool (static_cast<class LeafValue&>(*value));
    delete value;
    if (!rejected)
      return false;
  }
  if (state) {
    printer.printRaw (rejected
		      ? "rejected state "
		      : "fatally rejected state ");
    printer.printState (state);
  }
  else {
    printer.printRaw (rejected
		      ? "rejected "
		      : "fatally rejected ");
    m.display (printer);
  }
  printer.finish ();
  if (!rejected)
    return interrupted = true;
  return false;
}

bool
Net::computeInitMarking (const class Printer& printer)
{
  bool ok = true;
  assert (!myInitMarking);
  if (myInitMarking)
    return ok;
  myInitMarking = new class GlobalMarking (*this);
  class Valuation v;
  unsigned i;
  // non-implicit places
  for (i = 0; i < myNumPlaces; i++) {
    if (const class Marking* m = myPlaces[i]->getInitMarking ()) {
      if (!myPlaces[i]->isImplicit ()) {
	v.clearErrors ();
	if (!m->add ((*myInitMarking)[i], 1u, v)) {
	  assert (!v.isOK ());
	  printer.printRaw ("place ");
	  printer.print (myPlaces[i]->getName ());
	  printer.printRaw (": error in the initial marking expression:");
	  printer.finish ();
	  v.display (printer);
	  printer.finish ();
	  ok = false;
	}
      }
    }
  }
  // implicit places
  v.setGlobalMarking (myInitMarking);
  for (i = 0; i < myNumPlaces; i++) {
    if (const class Marking* m = myPlaces[i]->getInitMarking ()) {
      if (myPlaces[i]->isImplicit ()) {
	v.clearErrors ();
	if (!m->add ((*myInitMarking)[i], 1u, v)) {
	  assert (!v.isOK ());
	  printer.printRaw ("place ");
	  printer.print (myPlaces[i]->getName ());
	  printer.printRaw (": error in the initial marking expression:");
	  printer.finish ();
	  v.display (printer);
	  printer.finish ();
	  ok = false;
	}
      }
    }
  }
  if (isRejected (*myInitMarking, 0, printer))
    ok = false;
  return ok;
}

const class BoolType&
Net::getBoolType ()
{
  return boolType;
}

const class IntType&
Net::getIntType ()
{
  return intType;
}

const class CardType&
Net::getCardType ()
{
  return cardType;
}

const class CharType&
Net::getCharType ()
{
  return charType;
}

void
Net::analyze (const class Printer& printer,
	      class Graph& graph,
	      class SearchList& states,
	      enum SearchKind kind) const
{
  assert (kind < Single || states.size () == 1);

  while (!interrupted && !states.empty ()) {
    card_t source = states.pop (kind == Breadth);

    assert (!graph.isProcessed (source));

    class GlobalMarking* m = graph.fetchState (source);
    bool deadlock = true;

    for (unsigned i = 0; !interrupted && i < myNumTransitions; i++) {
      if (myTransitions[i]->analyze (*m, *this, graph, source,
				     kind == Single, states, printer)) {
	deadlock = false;
	if (myTransitions[i]->isPrioritized ())
	  break;
      }
    }

    graph.addEvents (source);

    if (!interrupted && deadlock && myDeadlock) {
      class Valuation valuation;
      valuation.setGlobalMarking (m);
      if (class Value* value = myDeadlock->eval (valuation)) {
	assert (value->getKind () == Value::vLeaf);
	deadlock = bool (static_cast<class LeafValue&>(*value));
	delete value;
	if (deadlock) {
	  printer.printRaw ("deadlock state ");
	  printer.printState (source);
	  printer.finish ();
	}
      }
      else {
	printer.printRaw ("fatal deadlock state ");
	printer.printState (source);
	printer.finish ();
	delete m;
	interrupted = true;
	return;
      }
    }

    delete m;

    if (kind >= Single)
      break;
  }
}

void
Net::analyze (const class Printer& printer,
	      class Graph& graph,
	      card_t state,
	      enum SearchKind kind) const
{
  if (!graph.isProcessed (state)) {
    class SearchList states;
    states.push (state);
    analyze (printer, graph, states, kind);
  }
}

void
Net::analyze (const class Printer& printer,
	      class HashGraph& graph,
	      bool breadth) const
{
  class StateList states;
  class BitPacker buf;
  if (!Graph::encode (buf, *myInitMarking, *myInitMarking, 0))
    return;
  buf.deflate ();
  graph.add (buf.getBuf (), buf.getNumBytes ());
  states.push (buf.getBuf (), buf.getNumBytes ());

  while (!interrupted && !states.empty ()) {
    class GlobalMarking m (*this);
    word_t* buf = states.pop (breadth);
    Graph::decode (m, *myInitMarking, buf);
    delete[] buf;
    bool deadlock = true;

    for (unsigned i = 0; !interrupted && i < myNumTransitions; i++) {
      if (myTransitions[i]->analyze (m, *this, graph, states, printer)) {
	deadlock = false;
	if (myTransitions[i]->isPrioritized ())
	  break;
      }
    }

    if (!interrupted && deadlock && myDeadlock) {
      class Valuation valuation;
      valuation.setGlobalMarking (&m);
      if (class Value* value = myDeadlock->eval (valuation)) {
	assert (value->getKind () == Value::vLeaf);
	deadlock = bool (static_cast<class LeafValue&>(*value));
	delete value;
	if (deadlock) {
	  printer.printRaw ("deadlock ");
	  m.display (printer);
	  printer.finish ();
	}
      }
      else {
	printer.printRaw ("fatal deadlock ");
	m.display (printer);
	printer.finish ();
	interrupted = true;
	return;
      }
    }
  }
}

bool
Net::isPredefinedType (const class Type& type)
{
  return ::isPredefinedType (&type);
}

#ifdef EXPR_COMPILE
# include "Compilation.h"
# include "CExpression.h"
# include "LSTS.h"

/** structure shared by analyze () and addState () for full analysis */
static struct
{
  /** source state number */
  card_t state;
  /** flag: deadlock state? */
  bool deadlock;
  /** flag: any successors generated? */
  bool successors;
  /** flag: add already generated successors to the list? */
  bool addOld;
  /** successor list */
  class SearchList* succ;
  /** the reachability graph */
  class Graph* graph;
  /** the printer object for diagnostics */
  const class Printer* printer;
  /** the compilation */
  const class Compilation* compilation;
}
context;

/** Add an encoded state to the reachability graph
 * @param buf	the encoded state
 * @param size	length of the encoded state, in bytes
 */
static void
addState (const void* buf, size_t size)
{
  Graph::AddStatus target = context.graph->add (buf, size);
  if (target.isNew)
    if (class LSTS* lsts = context.graph->getLSTS ())
      lsts->outputState (*context.compilation, target.number);
  context.graph->add (context.state, target.number);
  context.deadlock = false;
  context.successors = true;
  if (Net::theEdgeThresold &&
      !(context.graph->getNumArcs () % Net::theEdgeThresold)) {
    context.printer->print (card_t (context.graph->getNumStates ()));
    context.printer->printRaw (" states, ");
    context.printer->print (card_t (context.graph->getNumArcs ()));
    context.printer->printRaw (" arcs");
    context.printer->finish ();
  }
  if (target.isNew) {
    if (enum Error err = context.compilation->stateReject ()) {
      if (err != errComp) interrupted = true;
      context.printer->printRaw (err == errComp
				 ? "rejected state "
				 : "fatally rejected state ");
      context.printer->printState (target.number);
      context.printer->finish ();
    }
  }
  else if (!context.addOld)
    return;
  context.succ->push (target.number);
}

void
Net::analyze (const class Compilation& compilation,
	      const class Printer& printer,
	      class Graph& graph,
	      class SearchList& states,
	      enum SearchKind kind) const
{
  assert (kind < Single || states.size () == 1);
  context.addOld = kind == Single;
  context.succ = &states;
  context.graph = &graph;
  context.printer = &printer;
  context.compilation = &compilation;
  compilation.linkAddState (&::addState, true);

  while (!interrupted && !states.empty ()) {
    context.state = states.pop (kind == Breadth);
    context.deadlock = true;
    assert (!graph.isProcessed (context.state));
    graph.fetchState (compilation, context.state);
    compilation.eventClear ();
    for (unsigned tr = 0; !interrupted && tr < myNumTransitions; tr++) {
      context.successors = false;
      if (unsigned errors = compilation.eventAnalyze (tr)) {
	printer.printRaw ("state ");
	printer.printState (context.state);
	printer.printRaw (": transition ");
	printer.printQuoted (getTransition (tr).getName ());
	printer.delimiter (':');
	printer.print (errors);
	printer.printRaw (errors == 1 ? " error" : " errors");
	printer.finish ();
	graph.flagErroneous (context.state);
      }
      if (context.successors && myTransitions[tr]->isPrioritized ())
	break;
    }
    size_t bytes;
    const void* buf = compilation.eventDeflate (bytes);
    graph.addEvents (context.state, buf, bytes);

    if (enum Error err =
	context.deadlock ? compilation.stateDeadlock () : errNone) {
      printer.printRaw (err == errComp
			? "deadlock state "
			: "fatal deadlock state ");
      printer.printState (context.state);
      printer.finish ();
      if (err != errComp) {
	interrupted = true;
	return;
      }
    }

    if (kind >= Single)
      return;
  }
}

void
Net::analyze (const class Compilation& compilation,
	      const class Printer& printer,
	      class Graph& graph,
	      card_t state,
	      enum SearchKind kind) const
{
  if (!graph.isProcessed (state)) {
    class SearchList states;
    states.push (state);
    analyze (compilation, printer, graph, states, kind);
  }
}

/** structure shared by analyze () and probState () for probablistic method */
static struct
{
  /** flag: deadlock state? */
  bool deadlock;
  /** flag: any successors generated? */
  bool successors;
  /** successor list for probabilistic verification */
  class StateList* succ;
  /** probabilistic reachability set */
  class HashGraph* graph;
  /** the Petri net */
  const class Net* net;
  /** the printer object for diagnostics */
  const class Printer* printer;
  /** the compilation */
  const class Compilation* compilation;
} hashcontext;

/** Add an encoded state to the probabilistic reachability set
 * @param buf	the encoded state
 * @param size	length of the encoded state, in bytes
 */
static void
probState (const void* buf, size_t size)
{
  hashcontext.deadlock = false;
  hashcontext.successors = true;
  if (hashcontext.graph->add (buf, size)) {
    if (enum Error err = hashcontext.compilation->stateReject ()) {
      if (err != errComp) interrupted = true;
      hashcontext.printer->printRaw (err == errComp
				     ? "rejected "
				     : "fatally rejected ");
      size_t last = (size - 1) / sizeof (word_t);
      word_t* data = new word_t[last + 1];
      data[last] = 0;
      memcpy (data, buf, size);
      class GlobalMarking m (*hashcontext.net);
      Graph::decode (m, *hashcontext.net->getInitMarking (), data);
      delete[] data;
      m.display (*hashcontext.printer);
      hashcontext.printer->finish ();
    }
    hashcontext.succ->push (buf, size);
  }
  if (Net::theEdgeThresold &&
      !(hashcontext.graph->getNumArcs () % Net::theEdgeThresold)) {
    hashcontext.printer->print (hashcontext.graph->getNumStates ());
    hashcontext.printer->printRaw (" states, ");
    hashcontext.printer->print (hashcontext.graph->getNumArcs ());
    hashcontext.printer->printRaw (" arcs");
    hashcontext.printer->finish ();
  }
}

void
Net::analyze (const class Compilation& compilation,
	      const class Printer& printer,
	      class HashGraph& graph,
	      bool breadth) const
{
  class StateList states;
  class BitPacker buf;
  if (!Graph::encode (buf, *myInitMarking, *myInitMarking, 0))
    return;
  buf.deflate ();
  graph.add (buf.getBuf (), buf.getNumBytes ());
  states.push (buf.getBuf (), buf.getNumBytes ());

  hashcontext.succ = &states;
  hashcontext.graph = &graph;
  hashcontext.printer = &printer;
  hashcontext.net = this;
  hashcontext.compilation = &compilation;
  compilation.linkAddState (&::probState, false);

  while (!interrupted && !states.empty ()) {
    hashcontext.deadlock = true;
    size_t size;
    word_t* buf = states.pop (breadth, &size);
    compilation.stateDecode (buf, size);
    for (unsigned tr = myNumTransitions; !interrupted && tr--; ) {
      hashcontext.successors = false;
      if (unsigned errors = compilation.eventAnalyze (tr)) {
	class GlobalMarking m (*this);
	Graph::decode (m, *myInitMarking, buf);
	m.display (printer);
	printer.printRaw (" transition ");
	printer.printQuoted (getTransition (tr).getName ());
	printer.delimiter (':');
	printer.print (errors);
	printer.printRaw (errors == 1 ? " error" : " errors");
	printer.finish ();
      }
      if (hashcontext.successors && myTransitions[tr]->isPrioritized ())
	break;
    }

    if (enum Error err =
	hashcontext.deadlock ? compilation.stateDeadlock () : errNone) {
      printer.printRaw (err == errComp
			? "deadlock "
			: "fatal deadlock ");
      class GlobalMarking m (*this);
      Graph::decode (m, *myInitMarking, buf);
      m.display (printer);
      printer.finish ();
      if (err != errComp) {
	interrupted = true;
	delete[] buf;
	return;
      }
    }

    delete[] buf;
  }
}

void
Net::compileEncoder (class CExpression& cexpr) const
{
  unsigned i;
  class StringBuffer& out = cexpr.getOut ();
  class StringBuffer itemname;
  unsigned indent = 2;

  cexpr.setFatalError ("FREE(p); return *interrupted=1, errFatal;\n");

  /** flag: are there any implicit places? */
  bool hasImplicit = false;
  for (i = myNumPlaces; i--; ) {
    if (!myPlaces[i]->isImplicit ())
      continue;
    if (!hasImplicit) {
      hasImplicit = true;
      out.append ("#ifndef NO_INVARIANT_CHECK\n"
		  "# define NO_INVARIANT_CHECK(p) 0\n"
		  "#endif\n");
      out.indent (indent), out.append ("struct tree* p = 0;\n");
    }
    out.append ("#if !NO_INVARIANT_CHECK(");
    out.append (i), out.append (")\n");
    out.append ("# define ERROR(err){FREE(p);return err;}\n");
    if (myPlaces[i]->getMaxNumTokens () == 1) {
      out.indent (indent);
      out.append ("{\n");
      out.indent (indent += 2);
      myPlaces[i]->getType ().appendName (out);
      out.append ("* item;\n");
      myPlaces[i]->getInitMarking ()->compileScalarMset (cexpr, indent,
							 "item", 0, false);
      itemname.append ("(*");
      itemname.append (cexpr.getMultiset ());
      itemname.append (".p");
      itemname.append (i);
      itemname.append (")");
      if (!myPlaces[i]->getCapacityBits ()) {
	out.indent (indent);
	out.append ("if (!item)\n");
	cexpr.compileError (indent + 2, errConst);
      }
      out.indent (indent);
      out.append ("if (");
      if (myPlaces[i]->getCapacityBits ()) {
	out.append ("item != ");
	out.append (cexpr.getMultiset ());
	out.append (".p");
	out.append (i);
	out.append (" &&\n");
	out.indent (indent + 4);
      }
      if (!myPlaces[i]->getType ().
	  compileEqual (out, indent + 4, "(*item)", itemname.getString (),
			false, true, true, false))
	out.append ("0");
      out.append (")\n");
      cexpr.compileError (indent + 2, errComp);
      out.indent (indent -= 2);
      out.append ("}\n");
    }
    else {
      itemname.append ("(");
      myPlaces[i]->getType ().appendMSetName (itemname);
      itemname.append ("*)p");
      myPlaces[i]->getInitMarking ()->compileMset (cexpr, indent,
						   itemname.getString (), 0);
      out.indent (indent);
      out.append ("if (!equal");
      myPlaces[i]->getType ().appendIndex (out);
      out.append (" (");
      out.append (itemname);
      out.append (", ");
      out.append (cexpr.getMultiset ());
      out.append (".p");
      out.append (i);
      out.append ("))\n");
      cexpr.compileError (indent + 2, errComp);
    }
    out.append ("# undef ERROR\n"
		"  FREE (p);\n"
		"#endif /* NO_INVARIANT_CHECK */\n");
    cexpr.compileCleanup (2);
    itemname.create (0);
  }

  cexpr.setFatalError ("return *interrupted=1, errFatal;\n");
  out.indent (indent);
  out.append ("FCN (clear) ();\n");

  const char* total = cexpr.getVarTmpCount ();

  for (i = myNumPlaces; i--; ) {
    if (myPlaces[i]->isImplicit ())
      continue;
    out.append ("#define ERROR(err)return err\n");

    if (myPlaces[i]->getMaxNumTokens () == 1) {
      out.indent (indent);
      out.append ("if (!");
      out.append (cexpr.getMultiset ());
      out.append (".p");
      out.append (i);
      out.append (")\n");
      if (myPlaces[i]->getCapacityBits ()) {
	out.indent (indent + 2);
	out.append ("FCN (enc) (0, 1);\n");
	out.indent (indent);
	out.append ("else {\n");
	out.indent (indent += 2);
	out.append ("FCN (enc) (1, 1);\n");
      }
      else {
	cexpr.compileError (indent + 2, errConst);
      }

      out.indent (indent);
      out.append (total);
      out.append (" = 1;\n");

      itemname.append ("(*");
      itemname.append (cexpr.getMultiset ());
      itemname.append (".p");
      itemname.append (i);
      itemname.append (")");
      myPlaces[i]->getType ().compileEncoder (cexpr, indent, "FCN (enc)",
					      itemname.getString ());
      itemname.create (0);

      if (myPlaces[i]->getCapacityBits ()) {
	out.indent (indent -= 2);
	out.append ("}\n");
      }

      out.append ("#undef ERROR\n");
      continue;
    }

    out.indent (indent);
    out.append (total);
    out.append (" = msort (");
    out.append (cexpr.getMultiset ());
    out.append (".p");
    out.append (i);
    out.append (");\n");

    out.indent (indent);
    out.append ("if (");
    out.append (total);
    out.append (" == CARD_T_MAX)\n");
    cexpr.compileError (indent + 2, errCard);

    if (const class Constraint* c = myPlaces[i]->getCapacity ()) {
      c->compileCheck (cexpr, indent, total);
      if (unsigned cap = myPlaces[i]->getCapacityBits ()) {
	const char* number = cexpr.getVarCount ();
	if (cap > 2 && !myPlaces[i]->isNonempty ()) {
	  out.indent (indent);
	  out.append ("if ("), out.append (total), out.append (") {\n");
	  out.indent (indent + 2);
	  out.append ("FCN (enc) (1, 1);\n");
	  CardType::compileConversion (out, indent, *c, total, number);
	  out.indent (indent + 2);
	  out.append ("FCN (enc) ("), out.append (number), out.append (", ");
	  out.append (cap), out.append (");\n");
	  out.indent (indent);
	  out.append ("}\n");
	  out.indent (indent);
	  out.append ("else FCN (enc) (0, 1);\n");
	}
	else {
	  CardType::compileConversion (out, indent, *c, total, number);
	  out.indent (indent);
	  out.append ("FCN (enc) (");
	  out.append (number);
	  out.append (", ");
	  out.append (cap);
	  out.append (");\n");
	}
      }
    }
    else {
      out.indent (indent);
      out.append ("encc ("), out.append (total), out.append (");\n");
    }

    out.indent (indent);
    out.append ("if ("), out.append (total), out.append (") {\n");
    out.indent (indent + 2);
    out.append ("if ("), out.append (total);
    out.append (" > 1) FCN (enc) (distinct - 1, log2 (");
    out.append (total), out.append ("));\n");
    out.indent (indent + 2), out.append ("while (distinct--) {\n");
    out.indent (indent + 4), out.append ("register card_t low = (");
    out.append (total), out.append (" + distinct) / (distinct + 1);\n");
    out.indent (indent + 4), out.append ("register card_t high = ");
    out.append (total), out.append (" - distinct;\n");
    out.indent (indent + 4), out.append (total);
    out.append (" -= histogram[distinct]->count;\n");
    out.indent (indent + 4), out.append ("if (high != low)\n");
    out.indent (indent + 6);
    out.append ("FCN (enc) (histogram[distinct]->count - low, "
		"log2 (high - low + 1));\n");

    itemname.append ("((");
    myPlaces[i]->getType ().appendMSetName (itemname);
    itemname.append ("*)histogram[distinct])->item");
    myPlaces[i]->getType ().compileEncoder (cexpr, indent + 4, "FCN (enc)",
					    itemname.getString ());
    itemname.create (0);

    out.indent (indent + 2);
    out.append ("}\n");
    out.indent (indent);
    out.append ("}\n"
		"#undef ERROR\n");
  }

  out.append ("  FCN (deflate) ();\n"
	      "  (*addstate) (ENCVAR (buf), NUM_BYTES (ENCVAR (bits)));\n"
	      "  return errNone;\n");

  cexpr.setFatalError (0);
}

void
Net::compileDecoder (class CExpression& cexpr) const
{
  unsigned i;
  class StringBuffer& out = cexpr.getOut ();
  unsigned indent = 2;
  const char* total = cexpr.getVarTmpCount ();
  /** flag: are there any implicit places? */
  bool hasImplicit = false;
  class StringBuffer itemname;

  out.indent (indent);
  out.append ("freem (&");
  out.append (cexpr.getMultiset ());
  out.append (");\n");
  out.indent (indent);
  out.append ("FCN (inflate) ((word_t*) buf, size);\n");

  for (i = myNumPlaces; i--; ) {
    if (myPlaces[i]->isImplicit ()) {
      hasImplicit = true;
      continue;
    }
    if (myPlaces[i]->getMaxNumTokens () == 1) {
      out.indent (indent);
      if (myPlaces[i]->getCapacityBits ()) {
	out.append ("if ((numTokens["), out.append (i);
	out.append ("]=FCN (dec) (1))) {\n");
	out.indent (indent += 2);
      }
      else {
	out.append ("numTokens["), out.append (i), out.append ("]=1;\n");
	out.indent (indent);
      }
      out.append (cexpr.getMultiset ());
      out.append (".p");
      out.append (i);
      out.append (" = malloc (sizeof (");
      myPlaces[i]->getType ().appendName (out);
      out.append ("));\n");
      itemname.append ("(*");
      itemname.append (cexpr.getMultiset ());
      itemname.append (".p");
      itemname.append (i);
      itemname.append (")");
      myPlaces[i]->getType ().compileDecoder (cexpr, indent, "FCN (dec)",
					      itemname.getString ());
      if (myPlaces[i]->getCapacityBits ()) {
	out.indent (indent -= 2);
	out.append ("}\n");
      }
      itemname.create (0);
      continue;
    }
    else if (const class Constraint* c = myPlaces[i]->getCapacity ()) {
      if (unsigned cap = myPlaces[i]->getCapacityBits ()) {
	const char* number = cexpr.getVarCount ();
	if (cap > 2 && !myPlaces[i]->isNonempty ()) {
	  out.indent (indent), out.append ("if (FCN (dec) (1)) {\n");
	  out.indent (indent + 2);
	  out.append (number);
	  out.append (" = FCN (dec) ("), out.append (cap), out.append (");\n");
	  CardType::compileReverseConversion (out, indent + 2, *c,
					      number, total);
	  out.indent (indent), out.append ("}\n");
	  out.indent (indent), out.append ("else\n");
	  out.indent (indent + 2), out.append (total), out.append (" = 0;\n");
	}
	else {
	  out.indent (indent);
	  out.append (number);
	  out.append (" = FCN (dec) ("), out.append (cap), out.append (");\n");
	  CardType::compileReverseConversion (out, indent, *c,
					      number, total);
	}
      }
      else {
	out.indent (indent), out.append (total), out.append (" = ");
	out.append (CardType::convert (0, *c));
	out.append (";\n");
      }
    }
    else
      out.indent (indent), out.append (total), out.append (" = decc ();\n");

    out.indent (indent);
    out.append ("if ((numTokens["), out.append (i);
    out.append ("]="), out.append (total), out.append (")) {\n");
    out.indent (indent + 2);
    out.append ("for (distinct = "), out.append (total);
    out.append (" > 1 ? FCN (dec) (log2 ("), out.append (total);
    out.append (")) + 1 : 1;\n");
    out.indent (indent + 7);
    out.append ("distinct--; ) {\n");
    out.indent (indent + 4);
    myPlaces[i]->getType ().appendName (out);
    out.append (" item;\n");
    out.indent (indent + 4), out.append ("register card_t low = (");
    out.append (total), out.append (" + distinct) / (distinct + 1);\n");
    out.indent (indent + 4), out.append ("register card_t high = ");
    out.append (total), out.append (" - distinct;\n");
    out.indent (indent + 4);
    out.append ("register card_t amount = low != high\n");
    out.indent (indent + 6);
    out.append ("? low + FCN (dec) (log2 (high - low + 1))\n");
    out.indent (indent + 6);
    out.append (": low;\n");
    out.indent (indent + 4), out.append (total), out.append (" -= amount;\n");
    myPlaces[i]->getType ().compileDecoder (cexpr, indent + 4,
					    "FCN (dec)", "item");
    out.indent (indent + 4), out.append (cexpr.getMultiset ());
    out.append (".p"), out.append (i);
    out.append (" = insert"), myPlaces[i]->getType ().appendIndex (out);
    out.append (" ("), out.append (cexpr.getMultiset ());
    out.append (".p"), out.append (i);
    out.append (", &item, amount);\n");
    out.indent (indent + 2);
    out.append ("}\n");
    out.indent (indent);
    out.append ("}\n");
  }

  if (hasImplicit) {
    out.append ("#define ERROR(err) abort ()\n");
    for (i = myNumPlaces; i--; ) {
      if (!myPlaces[i]->isImplicit ())
	continue;
      itemname.append (cexpr.getMultiset ());
      itemname.append (".p"), itemname.append (i);
      if (myPlaces[i]->getMaxNumTokens () == 1) {
	myPlaces[i]->getInitMarking ()->
	  compileScalarMset (cexpr, indent, itemname.getString (), 0, false);
	out.indent (indent);
	out.append ("if ((numTokens["), out.append (i), out.append ("]=!!");
	out.append (itemname);
	out.append ("))\n");
	out.indent (indent + 2);
	out.append (itemname);
	out.append (" = memcpy (malloc (sizeof *");
	out.append (itemname);
	out.append ("), ");
	out.append (itemname);
	out.append (", sizeof *");
	out.append (itemname);
	out.append (");\n");
	if (!myPlaces[i]->getCapacityBits ()) {
	  out.indent (indent);
	  out.append ("else\n");
	  cexpr.compileError (indent + 2, errConst);
	}
      }
      else {
	myPlaces[i]->getInitMarking ()->
	  compileMset (cexpr, indent, itemname.getString (), 0);
	out.indent (indent), out.append ("{\n");
	out.indent (indent + 2), myPlaces[i]->getType ().appendMSetName (out);
	out.append ("* t = "), out.append (itemname), out.append (";\n");
	out.indent (indent + 2);
	out.append ("FIRST (t);\n");
	out.indent (indent + 2);
	out.append ("for (numTokens["), out.append (i), out.append ("]=0; ");
	out.append ("t; ) {\n");
	out.indent (indent + 4);
	out.append ("numTokens["), out.append (i), out.append ("] += ");
	out.append (itemname), out.append ("->count;\n");
	out.indent (indent + 4);
	out.append ("NEXT (t);\n");
	out.indent (indent + 2);
	out.append ("}\n");
	out.indent (indent);
	out.append ("}\n");
      }
      cexpr.compileCleanup (2);
      itemname.create (0);
    }
    out.append ("#undef ERROR\n");
  }
}

/** Compile a Boolean expression
 * @param cexpr		the compilation
 * @param expr		the expression to be compiled
 */
inline static void
compileExpression (class CExpression& cexpr,
		   const class Expression* expr)
{
  class StringBuffer& out = cexpr.getOut ();
  if (expr) {
    out.append ("#define ERROR(err)return err\n");
    const char* result = cexpr.getFlag ();
    expr->compile (cexpr, 2, result, 0);
    cexpr.compileCleanup (2);
    out.append ("#undef ERROR\n"
		"  if (");
    out.append (result);
    out.append (") return errComp;\n");
  }
  out.append ("  return errNone;\n");
}

void
Net::compileReject (class CExpression& cexpr) const
{
  ::compileExpression (cexpr, myReject);
}

void
Net::compileDeadlock (class CExpression& cexpr) const
{
  ::compileExpression (cexpr, myDeadlock);
}

void
Net::compileProps (class CExpression& cexpr,
		   const char* operation,
		   const char* data) const
{
  class StringBuffer& out = cexpr.getOut ();
  if (myNumProps) {
    out.append ("#define ERROR(err) continue\n");
    for (unsigned i = 0; i < myNumProps; i++) {
      out.append ("  do {\n");
      const char* result = cexpr.getFlag ();
      myProps[i]->compile (cexpr, 4, result, 0);
      out.append ("    if (");
      out.append (result);
      out.append (" && !");
      out.append (operation);
      out.append (" (");
      out.append (i);
      out.append (", ");
      out.append (data);
      out.append (")) return 0;\n");
      out.append ("  } while (0);\n");
    }
    cexpr.compileCleanup (2);
    out.append ("#undef ERROR\n");
  }
  out.append ("  return 1;\n");
}

void
Net::compileConstantDecl (class StringBuffer& out,
			  const char* ext) const
{
  class Printer printer;
  printer.setOutput (&out);
  for (unsigned i = 0; i < myNumConstants; i++) {
    if (ext)
      out.append (ext);
    else {
      out.append ("#if 0\n");
      printer.printQuoted (myConstantNames[i]);
      out.append ("\n"
		  "#endif\n");
    }
    myConstants[i]->getType ()->appendName (out);
    out.append (" c");
    out.append (i);
    out.append (";\n");
  }
}

void
Net::compileConstantInit (class StringBuffer& out) const
{
  char name[22];
  for (unsigned i = 0; i < myNumConstants; i++) {
    snprintf (name, sizeof name, "c%u", i);
    myConstants[i]->getValue ().compileInit (name, 2, out);
  }
}

#endif // EXPR_COMPILE

void
Net::display (const class Printer& printer) const
{
  for (TypeList::const_iterator t = myTypeList.begin ();
       t != myTypeList.end (); t++) {
    const class Type& type = **t;
    if (!type.getName ()) continue;
    printer.printRaw ("typedef ");
    type.display (printer);
    printer.delimiter (' ');
    printer.print (type.getName ());
    printer.delimiter (';');
    printer.finish ();
  }

  for (unsigned p = 0; p < myNumPlaces; p++)
    myPlaces[p]->display (printer);

  for (unsigned tr = 0; tr < myNumTransitions; tr++)
    myTransitions[tr]->display (printer);

  if (myReject) {
    printer.printRaw ("reject ");
    myReject->display (printer);
    printer.delimiter (';');
    printer.finish ();
  }
  if (myDeadlock) {
    printer.printRaw ("deadlock ");
    myDeadlock->display (printer);
    printer.delimiter (';');
    printer.finish ();
  }

  for (unsigned prop = 0; prop < myNumProps; prop++) {
    printer.printRaw ("prop ");
    printer.print (myPropNames[prop]);
    printer.delimiter (':')++;
    myProps[prop]->display (printer);
    printer--;
    printer.delimiter (';');
    printer.finish ();
  }
}
