#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <clocale>

#ifdef __alpha
# include </usr/include/signal.h>
# include <sys/errno.h>
#else // __alpha
# include <csignal>
#endif // __alpha

#include <cerrno>
#include <sys/stat.h>
#include <sys/types.h>
#include <time.h>
#ifdef __unix
# include <sys/times.h>
# include <sys/wait.h>
#endif // __unix
#ifdef __WIN32
# undef __STRICT_ANSI__
# include <windows.h>
# include <io.h>
# undef mkdir
# define mkdir(name,perm) _mkdir (name)
# include <libiberty/getopt.h>
# ifndef HAS_GETOPT_LONG
#  define HAS_GETOPT_LONG
# endif // HAS_GETOPT_LONG
#elif defined HAS_GETOPT_LONG
# include <getopt.h>
#endif // __WIN32

#ifdef HAS_REGEX
# include <regex.h>
#endif // HAS_REGEX

#ifdef EXPR_COMPILE
# include "Compilation.h"
#endif // EXPR_COMPILE

#include <unistd.h>

#include "version.h"
#include "Dotty.h"

#include "Expression.h"
#include "Valuation.h"
#include "GlobalMarking.h"
#include "Transition.h"

/** @mainpage Maria: Modular Reachability Analyser for Algebraic System Nets
 * @htmlonly
 * Please visit also the
 * <a href="http://www.tcs.hut.fi/maria/">Maria home page</a>.
 * @endhtmlonly
 */

/**
 * @file maria.C
 * The main program of Maria
 */

#ifdef HAS_GETOPT_LONG
/** A unified interface to getopt(3) and the GNU variant getopt_long(3) */
# define getopt(argc,argv,optstring) getopt_long(argc,argv,optstring,opts,0)
static const struct option opts[] = {
  { "include", 1, 0, 'I' },
  { "define", 1, 0, 'D' },
  { "unfold", 1, 0, 'u' },
  { "undefine", 1, 0, 'U' },
  { "help", 0, 0, '?' },
  { "verbose", 0, 0, 'v' },
  { "version", 0, 0, 'V' },
  { "warnings", 0, 0, 'W' },
  { "no-warnings", 0, 0, 'w' },
  { "property-translator", 1, 0, 'p' },
  { "array-limit", 1, 0, 'a' },
  { "quantification-limit", 1, 0, 'q' },
  { "breadth-first-search", 1, 0, 'b' },
  { "depth-first-search", 1, 0, 'd' },
  { "model", 1, 0, 'm' },
  { "probabilistic", 1, 0, 'P' },
  { "hashes", 1, 0, 'H' },
  { "graph", 1, 0, 'g' },
  { "execute", 1, 0, 'e' },
  { "edges", 1, 0, 'E' },
  { "width", 1, 0, 'i' },
  { "radix", 1, 0, 'x' },
# ifdef HAS_REGEX
  { "name", 1, 0, 'N' },
  { "no-name", 1, 0, 'n' },
# endif // HAS_REGEX
# ifdef EXPR_COMPILE
  { "compile", 1, 0, 'C' },
  { "no-compile", 0, 0, 'c' },
# endif // EXPR_COMPILE
  { 0, 0, 0, 0 }
};
#endif // HAS_GETOPT_LONG

/** Flag: be verbose */
static bool verbose = false;
/** Translator for property automata */
char* translator = NULL;
#ifdef EXPR_COMPILE
/** name of the compilation directory (NULL=don't compile) */
char* compiledir = 0;
/** The compiled net */
static class Compilation* compilation = 0;
/** Command-line flags for compiled expressions */
# define EXPRFLAGS "cC:"
#else
/** Command-line flags for compiled expressions */
# define EXPRFLAGS ""
#endif // EXPR_COMPILE

#ifdef HAS_REGEX
/** regular expression identifiers */
static const char regexps[] = {
  'P', // places
  'T', // transitions
  't', // types
  'e', // enumeration names
  'c', // struct or union components
  'f', // functions
  'p', // parameters
  'v', // variables
  'i', // iterator variables
};
/** regular expressions for allowed names */
static regex_t* regexpAllow[sizeof regexps] = { 0, };
/** regular expressions for disallowed names */
static regex_t* regexpDisallow[sizeof regexps] = { 0, };
/** extract regular expressions from a string
 * @param s	a regular expression prefixed with a character
 * @param regex	compiled regular expressions
 */
static void regexpExtract (const char* s, regex_t** regex);
/** deallocate compiled regular expressions
 * @param regex		the compiled regular expressions
 */
static void regexpFree (regex_t** regex);
/** Command-line flags for specifying regular expressions */
# define REGEXFLAGS "n:N:"
#else // HAS_REGEX
/** Command-line flags for specifying regular expressions */
# define REGEXFLAGS ""
#endif // HAS_REGEX

/** Option string for getopt(3) */
static const
char optstring[] =
"I:D:u:U:?hvVwWp:a:q:b:d:m:P:H:g:e:E:i:x:" EXPRFLAGS REGEXFLAGS;

#include "util.h"
#include "NameList.h"
static class NameList includePath;

#include "Printer.h"
#include "Net.h"
#include "Graph.h"
#include "ComponentGraph.h"
#include "LSTS.h"
#include "Property.h"
#include "Product.h"
#include "LNet.h"
#include "HashGraph.h"

/** The Bison-generated Petri Net parser
 * @return	zero if exiting due to end-of-file; 1 if parse error occurred
 */
extern int pnparse ();
/** message prefix function for the Petri Net parser
 * @return	number of characters written
 */
extern unsigned pnmsg (void);

/** The Bison-generated scripting language parser
 * @return	zero if exiting due to end-of-file; 1 if parse error occurred
 */
extern int deparse ();
/** Read a command line */
extern const char* deline ();
/** Reset the scripting language parser */
extern void dereset ();
/** message prefix function for the script parser
 * @return	number of characters written
 */
extern unsigned demsg (void);

/** Define a preprocessor symbol
 * @param sym	name of the symbol
 */
extern void define_sym (const char* sym);
/** Undefine a preprocessor symbol
 * @param sym	name of the symbol
 */
extern void undef_sym (const char* sym);

/** warning flag (true=print warnings) */
extern bool pnwarnings;
/** Maximum index type size of arrays */
extern card_t maxIndex;
/** Maximum range size of quantifications */
extern card_t maxRange;

/** The printer object */
class Printer thePrinter;
/** The Petri Net being parsed */
class Net* net = 0;
/** The reachability graph */
static class Graph* graph = 0;
/** The strongly connected component graph of the reachability graph */
static class ComponentGraph* componentgraph = 0;
/** Name of the Petri Net being parsed */
const char* netname;
/** Number of elements in the hash tables in probabilistic analysis */
static size_t hashSize = 0x7ffff;
/** Number of entries in the hash functions in probabilistic analysis */
static size_t funcSize = 16;
/** Number of hash tables in probabilistic analysis */
static size_t numTables = 2;

/** Name of the script being parsed */
const char* scriptname;
/** Name of scripts parsed from the command line */
const char* scriptcommand = "(command line)";
/** The script to be parsed from the command line */
const char* scriptstring;

/** Name of the program */
static const char* progname;

/** Check the validity of an identifier
 * @param name		name of the identifier
 * @param type		type of the identifier
 * @param msg		error message
 */
void checkName (const char* name, unsigned type, const char* msg);

/** Open a model
 * @param filename	name of the net to be analyzed
 * @param graphfile	base name for the reachability graph files
 * @param regen		flag: regenerate the graph files
 *			(if !graphfile && !regen, generate no graph files)
 * @return		false on error, true otherwise
 */
bool model (const char* filename, const char* graphfile, bool regen);

/** Perform on-the-fly model checking
 * @param state		state to start the analysis with
 * @param expr		formula to be verified
 * @param visual	visualization level
 */
static void onthefly (card_t state, class Expression* expr, card_t visual);

/** Generate an LSTS of the model when generating the state space
 * @param filebase	base file name (NULL=disable LSTS generation)
 */
void lsts (const char* filebase);

/** Unfold the model
 * @param ffilename	output format and file name (NULL=default,stdout)
 */
void unfold (const char* ffilename);

/** Dump the syntax tree of the model
 * @param visual	visualization level
 */
void dump (card_t visual);

/** Perform exhaustive reachability analysis
 * @param state		state to start the analysis with
 * @param breadth	flag: go breadth-first instead of depth-first
 */
void analyze (card_t state, bool breadth);

/** Open an input file
 * @param name		name of the input file
 * @return		a descriptor to the file, or NULL on error
 */
FILE* openfile (const char* name);
/** Open a script
 * @param name		name of the script
 * @return		a descriptor to the file, or NULL on error
 */
FILE* openscript (const char* name);

/** Execute a script
 * @param filename	name of the script to be executed
 * @return		false on a fatal error, true otherwise
 */
static bool execute (const char* filename);
/** Display a help message of the command line options */
static void cmdhelp (void);
/** Check if a state number exists
 * @param state	the state
 * @return	true if the state exists
 */
bool isState (card_t state);
/** Show a state
 * @param state		the state
 * @param visual	visualization level
 */
void show (card_t state, card_t visual);
/** Show a path specified as a state list
 * @param sl		the state list
 * @param visual	visualization level
 */
void show (const card_t* states, card_t visual);
/** Show the successors of a state
 * @param state		the state
 * @param visual	visualization level
 */
void succ (card_t state, card_t visual);
/** Show the predecessors of a state
 * @param state		the state
 * @param visual	visualization level
 */
void pred (card_t state, card_t visual);
/** Display statistics */
void stats (void);

/** Time used by the program since the last measurement */
static clock_t lastclock;

/** Display execution times */
void showtimes (void);
/** Compute strongly connected components
 * @param state	state in which to start the search
 * @param cond	condition for the states to be included in the search
 */
void strong (card_t state, class Expression* cond);
/** Display non-trivial terminal strongly connected components */
void terminal (void);
/** Display all strongly connected components
 * @param visual	visualization level
 */
void allcomps (card_t visual);
/** Show the nodes of a strongly connected component fulfilling a condition
 * @param comp		the component
 * @param cond		the condition (NULL=true)
 * @param visual	visualization level
 */
void comp_show (card_t comp,
		class Expression* cond,
		card_t visual);

/** Show the successors of a strongly connected component
 * @param comp		the component
 * @param visual	visualization level
 */
void comp_succ (card_t comp, card_t visual);
/** Show the predecessors of a strongly connected component
 * @param comp		the component
 * @param visual	visualization level
 */
void comp_pred (card_t comp, card_t visual);
/** Display a labelled state transition system of a path
 * @param lsts		the output system
 * @param path		the path
 * @param reverse	flag: display the path in reverse order?
 */
static void
printLSTS (class LSTS& lsts,
	   const card_t* path,
	   bool reverse);
/** Display a path
 * @param path	the path
 */
static void printPath (const card_t* path);
/** Display a path in reverse order
 * @param path	the path
 */
static void printReversePath (const card_t* path);
/** Show the shortest path from a state to a strongly connected component
 * @param state		the state
 * @param comp		the component
 * @param pathc		path condition (must hold in every state on the path)
 * @param visual	visualization level
 */
void path_sc (card_t state, card_t comp,
	      class Expression* pathc,
	      card_t visual);
/** Show the shortest path from a strongly connected component to a state
 * @param comp		the component
 * @param state		the state
 * @param pathc		path condition (must hold in every state on the path)
 * @param visual	visualization level
 */
void path_cs (card_t comp, card_t state,
	      class Expression* pathc,
	      card_t visual);
/** Show the shortest path from a state to a state
 * @param state		the source state
 * @param dest		the target state
 * @param pathc		path condition (must hold in every state on the path)
 * @param visual	visualization level
 */
void path_ss (card_t state, card_t dest,
	      class Expression* pathc,
	      card_t visual);
/** Show the shortest path from a state to a state where a condition holds
 * @param state		the source state
 * @param cond		the condition
 * @param pathc		path condition (must hold in every state on the path)
 * @param visual	visualization level
 */
void path_se (card_t state, class Expression& cond,
	      class Expression* pathc,
	      card_t visual);
/** Show the shortest path from a state where a condition holds to a state
 * @param state		the target state
 * @param cond		the condition
 * @param pathc		path condition (must hold in every state on the path)
 * @param visual	visualization level
 */
void rpath_se (card_t state, class Expression& cond,
	       class Expression* pathc,
	       card_t visual);
/** Show the shortest path from a state to a loop state, plus the loop
 * @param state		the source state
 * @param loop		amount and number of states in the loop
 * @param pathc		path condition (must hold in every state on the path)
 * @param visual	visualization level
 */
void path_sl (card_t state,
	      const card_t* loop,
	      class Expression* pathc,
	      card_t visual);

/** Evaluate a formula in the current state
 * @param state		state in which to evaluate the formula
 * @param expr		formula to be evaluated
 * @param visual	visualization level for counterexample
 */
void eval (card_t state, class Expression* expr, card_t visual);

/** Flag: has the analysis been interrupted? */
extern volatile bool interrupted;
#if defined __WIN32
/** Visualization process handle */
extern volatile HANDLE vispid;
#elif defined __unix
/** Visualization process identifier */
extern volatile pid_t vispid;
#endif // __unix
/** The visualization object */
static class Dotty* dotty;
/** Initialize the visualization process and invoke one of its methods
 * @param x	the method to be called
 */
#define vis(x) do { if (initDotty ()) dotty->x; } while (0)

/** Initialize the visualization interface
 * @return	pointer to the visualization object, or NULL on failure
 */
static const class Dotty*
initDotty (void)
{
  if (!dotty) {
    size_t len = strlen (progname);
    char* visname = new char[len + 5];
    memcpy (visname, progname, len);
    memcpy (visname + len, "-vis", 5);
    dotty = Dotty::startVisual (visname);
    delete[] visname;
    if (!dotty)
      return 0;
  }

  dotty->setParams (*graph, componentgraph,
#ifdef EXPR_COMPILE
		    compilation,
#endif // EXPR_COMPILE
		    thePrinter);
  return dotty;
}

#ifdef __unix
/** Handle a child termination
 * @param pid		process number of the terminated child
 */
extern "C"
void
childterm (pid_t pid)
{
  if (dotty && pid == vispid) {
    dotty->stopVisual ();
    dotty = 0;
  }
}
#endif // __unix

/** Handle a signal
 * @param num		number of the signal
 */
extern "C"
void
sig (int num) {
  if (num == SIGINT)
    interrupted = true;
#ifdef __unix
  else if (num == SIGCHLD)
    childterm (wait (0));
#endif // __unix
  signal (num, sig);
}

/** Clean everything up at program exit */
extern "C"
void
cleanup (void) {
#ifdef HAS_REGEX
  regexpFree (regexpAllow);
  regexpFree (regexpDisallow);
#endif // HAS_REGEX
#ifdef EXPR_COMPILE
  delete compilation;
  delete[] compiledir;
#endif // EXPR_COMPILE
  if (dotty) {
#if defined __WIN32
    TerminateProcess (vispid, 0);
#elif defined __unix
    kill (vispid, SIGKILL);
#endif // __unix
    dotty->stopVisual ();
  }
  delete net;
  delete graph;
  delete componentgraph;
  delete[] translator;
}

/** The main function
 * @param argc	argument count
 * @param argv	argument vector
 * @return	0 if successful;
 *		1 on parse error
 */
int
main (int argc, char** argv)
{
#ifndef __unix
  lastclock = clock ();
#else // !__unix
  struct tms current;
  lastclock = times (&current);
#endif // !__unix
  setlocale (LC_ALL, "");
  signal (SIGINT, sig);
#ifdef __unix
  signal (SIGCHLD, sig);
#endif // __unix
  atexit (cleanup);
  progname = *argv;

  /** flag: has the computation been aborted? */
  extern bool exiting;

#if defined YYDEBUG && YYDEBUG
  {
    /** Debugging flag for the Bison-generated parser */
    extern int pndebug;
    char *debug = getenv ("DEBUG");
    if (debug) pndebug = atoi (debug);
  }
#endif // YYDEBUG

  for (;;) {
    int c = getopt (argc, argv, optstring);
    if (c == EOF)
      break;
    switch (c) {
    case ':':
      // fall through
    default:
    case '?':
    case 'h':
      cmdhelp ();
      return 0;
    case 'v':
      verbose = true;
      break;
    case 'V':
      fprintf (stderr, "%s version " VERSION "\n", progname);
      return 0;
    case 'w':
      pnwarnings = false;
      break;
    case 'W':
      pnwarnings = true;
      break;
    case 'd':
      if (model (optarg, 0, true))
	analyze (0, false);
      break;
    case 'b':
      if (model (optarg, 0, true))
	analyze (0, true);
      break;
    case 'g':
      model (0, optarg, false);
      break;
    case 'm':
      model (optarg, 0, true);
      break;
    case 'P':
      model (optarg, 0, false);
      break;
    case 'H':
      if (sscanf (optarg, "%u%n,%u%n,%u %n",
		  &hashSize, &c, &funcSize, &c, &numTables, &c) < 1 ||
	  optarg[c] ||
	  !hashSize || !funcSize || !numTables) {
	fprintf (stderr, "error in hash parameters: `%s'\n", optarg);
	return 1;
      }
    case 'p':
      delete[] translator;
      translator = newString (optarg);
      break;
    case 'a':
      {
	char* end = NULL;
	maxIndex = strtoul (optarg, &end, 0);
	if (!*optarg || *end) {
	  fprintf (stderr, "error in array size limit: `%s'\n", optarg);
	  return 1;
	}
      }
      break;
    case 'q':
      {
	char* end = NULL;
	maxRange = strtoul (optarg, &end, 0);
	if (!*optarg || *end) {
	  fprintf (stderr, "error in quantification limit: `%s'\n", optarg);
	  return 1;
	}
      }
      break;
    case 'i':
      {
	char* end = NULL;
	unsigned width = strtoul (optarg, &end, 0);
	if (!*optarg || *end) {
	  fprintf (stderr, "error in output width limit: `%s'\n", optarg);
	  return 1;
	}
	else
	  thePrinter.setWidth (width);
      }
      break;
    case 'x':
      {
	if (!strcmp (optarg, "oct") || !strcmp (optarg, "octal") ||
	    !strcmp (optarg, "8"))
	  thePrinter.setRadix (Printer::Octal);
	else if (!strcmp (optarg, "hex") || !strcmp (optarg, "hexadecimal") ||
		 !strcmp (optarg, "16"))
	  thePrinter.setRadix (Printer::Hexadecimal);
	else if (!strcmp (optarg, "dec") || !strcmp (optarg, "decimal") ||
		 !strcmp (optarg, "10"))
	  thePrinter.setRadix (Printer::Decimal);
	else
	  fprintf (stderr, "ignoring unknown radix setting: `%s'\n", optarg);
      }
      break;
    case 'I':
      includePath.append (newString (optarg));
      break;
    case 'u':
      unfold (optarg);
      break;
    case 'D':
      define_sym (optarg);
      break;
    case 'U':
      undef_sym (optarg);
      break;
    case 'e':
      if ((scriptstring = optarg) &&
	  (interrupted || !execute (scriptcommand) || interrupted))
	return 4;
      if (exiting)
	return 0;
      break;
    case 'E':
      {
	char* end = NULL;
	Net::theEdgeThresold = strtoul (optarg, &end, 0);
	if (!*optarg || *end) {
	  fprintf (stderr, "error in edge thresold: `%s'\n", optarg);
	  return 1;
	}
      }
      break;
#ifdef HAS_REGEX
    case 'n':
      regexpExtract (optarg, regexpDisallow);
      break;
    case 'N':
      regexpExtract (optarg, regexpAllow);
      break;
#endif // HAS_REGEX
#ifdef EXPR_COMPILE
    case 'c':
      delete[] compiledir;
      compiledir = 0;
      break;
    case 'C':
      delete[] compiledir;
      compiledir = newString (optarg);
      break;
#endif // EXPR_COMPILE
    }
  }

  for (int i = optind; i < argc; i++)
    if (exiting)
      return 0;
    else if (interrupted || !execute (argv[i]) || interrupted)
      return 4;

  dereset ();
  while (!exiting && (scriptstring = deline ()))
    execute (scriptcommand);

  return 0;
}

static bool
execute (const char* filename)
{
  /** flag: has the computation been aborted? */
  extern bool exiting;
  scriptname = filename;
  thePrinter.setBOL (scriptname == scriptcommand ? 0 : &demsg);
  dereset ();
  bool status = !deparse () || exiting;
  thePrinter.setBOL (0);
  return status;
}

#ifdef HAS_REGEX
static void
regexpExtract (const char* s, regex_t** regex)
{
  unsigned index;
  for (index = 0; index < sizeof regexps; index++)
    if (*s == regexps[index])
      break;
  if (index < sizeof regexps) {
    if (regex[index]) {
      regfree (regex[index]);
      regex[index] = 0;
    }
    if (int errcode = regcomp (regex[index] = new regex_t,
			       &s[1], REG_EXTENDED | REG_NOSUB)) {
      size_t size = regerror (errcode, regex[index], 0, 0);
      char* msg = new char[size];
      regerror (errcode, regex[index], msg, size);
      fprintf (stderr, "%s:invalid regular expression '%c' \"%s\": %s\n",
	       progname, *s, &s[1], msg);
      delete[] msg;
      delete regex[index];
      regex[index] = 0;
    }
  }
  else
    fprintf (stderr, "%s:unknown regular expression '%c'\n", progname, *s);
}

static void
regexpFree (regex_t** regex)
{
  for (unsigned index = sizeof regexps; index--; ) {
    if (regex[index]) {
      regfree (regex[index]);
      delete regex[index];
    }
  }
}

void
checkName (const char* name, unsigned type, const char* msg)
{
  assert (name && type < sizeof regexps && msg);
  if (regexpAllow[type] &&
      !regexec (regexpAllow[type], name, 0, 0, 0))
    return;
  if (regexpDisallow[type] &&
      !regexec (regexpDisallow[type], name, 0, 0, 0)) {
    thePrinter.printRaw (msg);
    thePrinter.print (name);
    thePrinter.finish ();
    /** number of errors encountered */
    extern unsigned pnerrors;
    pnerrors++;
  }
}
#else // HAS_REGEX
void
checkName (const char*, unsigned, const char*)
{
}
#endif // HAS_REGEX

static void
cmdhelp (void)
{
  fprintf (stderr, "Usage: %s [options] scriptfile...\n", progname);
  fputs ("-a\t--array-limit LIMIT\n"
	 "-b\t--breadth-first-search MODELNAME\n"
#ifdef EXPR_COMPILE
	 "-C\t--compile DIRECTORY\n"
	 "-c\t--no-compile\n"
#endif // EXPR_COMPILE
	 "-D\t--define SYMBOL\n"
	 "-d\t--depth-first-search MODELNAME\n"
	 "-E\t--edges number\n"
	 "-e\t--execute SCRIPT\n"
	 "-g\t--graph GRAPHFILE\n"
	 "-H\t--hashes h[,f[,t]]\n"
	 "-h\t--help\n"
	 "-I\t--include DIRECTORY\n"
	 "-i\t--width columns\n"
	 "-m\t--model MODELNAME\n"
#ifdef HAS_REGEX
	 "-N\t--name [PTtecfpvi]RE\n"
	 "-n\t--no-name [PTtecfpvi]RE\n"
#endif // HAS_REGEX
	 "-P\t--probabilistic MODELNAME\n"
	 "-p\t--property-translator COMMAND\n"
	 "-q\t--quantification-limit LIMIT\n"
	 "-u\t--unfold M?([lmpr](OUTFILE)?)?\n"
	 "-U\t--undefine SYMBOL\n"
	 "-v\t--verbose\n"
	 "-V\t--version\n"
	 "-W\t--warnings\n"
	 "-w\t--no-warnings\n"
	 "-x\t--radix BASE\n",
	 stderr);
}

bool
model (const char* filename, const char* graphfile, bool regen)
{
  assert (filename || graphfile);
  thePrinter.setBOL (&pnmsg);
#ifdef EXPR_COMPILE
  delete compilation;
  compilation = 0;
#endif // EXPR_COMPILE
  delete net;
  net = new class Net;
  delete graph;
  graph = 0;
  delete componentgraph;
  componentgraph = 0;

  if (graphfile || regen) {
    graph = new class Graph (*net, filename, graphfile);
    if (!graph->openFiles (regen) && !filename) {
      fprintf (stderr, "%s:could not open the graph files\n", graphfile);
      delete net;
      net = 0;
      delete graph;
      graph = 0;
      return false;
    }
    if (!filename)
      filename = graph->getFilename ();
  }

  netname = filename;

  /** number of errors encountered */
  extern unsigned pnerrors;
  pnerrors = 0;
  if (pnparse () || pnerrors) {
    fprintf (stderr, "%s:%u %s\n", netname, pnerrors,
	     pnerrors == 1 ? "error" : "errors");
    delete net;
    net = 0;
    delete graph;
    graph = 0;
    thePrinter.setBOL (0);
    return false;
  }

  thePrinter.setBOL (0);

  if (verbose) {
    fprintf (stderr, "%s:%u places and %u transitions\n",
	     netname, net->getNumPlaces (), net->getNumTransitions ());
    fprintf (stderr, "%s:computing the initial marking\n", netname);
  }
  if (!net->computeInitMarking (thePrinter))
    fprintf (stderr, "%s:error in the initial marking\n", netname);
  else {
    class BitPacker buf;
    class Graph::AddStatus status (false, 0);
    if (!Graph::encode (buf, *net->getInitMarking (),
			*net->getInitMarking (), 0))
      fprintf (stderr, "%s:invalid initial marking\n", netname);
    else if (graph && (status = graph->add (*net->getInitMarking (), 0),
		       !status.isNew && status.number == CARD_T_MAX))
      fprintf (stderr, "%s:invalid initial marking\n", netname);
    else if (graph && status.number != 0)
      fprintf (stderr, "%s:inconsistent reachability graph file\n", netname);
#ifdef EXPR_COMPILE
    else if (compiledir) {
      if (verbose)
	fprintf (stderr, "%s:compiling\n", netname);
      compilation = new class Compilation (*net, compiledir, progname);
      if (!compilation->compile ())
	fprintf (stderr, "%s:compilation failed\n", netname);
      else if (char* errmsg = compilation->link ()) {
	fprintf (stderr, "%s:%s\n", netname, errmsg);
	delete[] errmsg;
      }
      else
	return true;
    }
#endif // EXPR_COMPILE
    else
      return true;
  }

#ifdef EXPR_COMPILE
  delete compilation;
  compilation = 0;
#endif // EXPR_COMPILE
  delete graph;
  graph = 0;
  delete net;
  net = 0;
  return false;
}

static void
onthefly (card_t state, class Expression* expr, card_t visual)
{
  if (!isState (state));
  else if (!translator || !*translator) {
    thePrinter.printRaw ("no translator for property automata defined");
    thePrinter.finish ();
  }
  else {
    if (verbose)
      fprintf (stderr, "%s:translating the formula\n", netname);
    class Property property (thePrinter);
    if (!property.create (translator, *expr)) {
      thePrinter.printRaw ("could not translate the formula");
      thePrinter.finish ();
    }
    else {
      class Product product (*graph, property);
      extern volatile bool interrupted;
      card_t* counterexample;
#ifdef EXPR_COMPILE
      if (compilation) {
	if (verbose)
	  fprintf (stderr, "%s:compiling the property automaton\n", netname);
	if (!compilation->compile (property)) {
	  thePrinter.printRaw ("could not compile the property automaton");
	  thePrinter.finish ();
	  return;
	}
	if (verbose)
	  fprintf (stderr, "%s:performing model checking\n", netname);
	counterexample = product.analyze (*compilation, state);
      }
      else {
#endif /* EXPR_COMPILE */
	if (verbose)
	  fprintf (stderr, "%s:performing model checking\n", netname);
	counterexample = product.analyze (state);
#ifdef EXPR_COMPILE
      }
#endif /* EXPR_COMPILE */

      if (!interrupted) {
	if (!counterexample) {
	  thePrinter.printRaw ("property holds");
	  thePrinter.finish ();
	}
 	else if (visual)
 	  vis (show (counterexample, visual > 1, true));
 	else {
 	  thePrinter.printRaw ("counterexample path:");
 	  thePrinter++.linebreak ();
 	  printReversePath (counterexample);
 	  thePrinter--;
	  thePrinter.finish ();
  	}
      }

      delete[] counterexample;
    }
  }
}

void
lsts (const char* filebase)
{
  if (!filebase) {
    if (graph) graph->setLSTS (0);
    return;
  }
  if (!isState (0))
    return;
  class LSTS* l = new class LSTS (*net, filebase);
  if (l->openFiles ()) {
    graph->setLSTS (l);
    l->outputState (*net->getInitMarking (), 0);
  }
  else
    delete l;
}

void
unfold (const char* ffilename)
{
  /** output format */
  enum { fmtDefault, fmtLoLA, fmtPEP, fmtPROD } format = fmtDefault;
  /** flag: generate a minimal unfolding */
  bool minimal = false;
  /** primary and secondary output file */
  FILE* f = stdout, *f2 = stdout;

  if (!net) {
    thePrinter.printRaw ("no model has been loaded");
    thePrinter.finish ();
    return;
  }

  if (ffilename) {
    if (*ffilename == 'M') minimal = true, ffilename++;
    switch (*ffilename) {
    case 'l':
      format = fmtLoLA; break;
    case 'p':
      format = fmtPEP; break;
    case 'r':
      format = fmtPROD; break;
    case 'm': case 0:
      break;
    default:
      thePrinter.printRaw ("unknown output format ");
      thePrinter.print (char_t (*ffilename));
      thePrinter.printRaw (", using default");
      thePrinter.finish ();
    }
    if (*ffilename && *++ffilename) {
      if (format == fmtPROD) {
	const char* suffix;
	if (!(suffix = strrchr (ffilename, '.')) ||
	    strcmp (suffix, ".net")) {
	  thePrinter.printRaw ("PROD model name ");
	  thePrinter.printQuoted (ffilename);
	  thePrinter.printRaw (" does not end in \".net\"");
	  thePrinter.finish ();
	  return;
	}
	static const char tbl[] = ".src/tables.c";
	char* tablename = new char[suffix - ffilename + sizeof tbl];
	memcpy (tablename, ffilename, suffix - ffilename);
	memcpy (tablename + (suffix - ffilename), tbl, sizeof tbl);
	if (!(f2 = fopen (tablename, "w"))) {
	  if (errno != ENOENT) {
	    delete[] tablename;
	    perror ("fopen");
	    return;
	  }
	  tablename[suffix - ffilename + 4] = 0;
	  if (mkdir (tablename, 0777)) {
	    delete[] tablename;
	    perror ("mkdir");
	    return;
	  }
	  tablename[suffix - ffilename + 4] = '/';
	  if (!(f2 = fopen (tablename, "w"))) {
	    delete[] tablename;
	    perror ("fopen");
	    return;
	  }
	}
      }

      if (!(f = fopen (ffilename, "w"))) {
	perror ("fopen");
	if (f2 != stdout)
	  fclose (f2);
	return;
      }
    }
  }

  if (verbose)
    fprintf (stderr, "%s:unfolding\n", netname);

  class LNet lnet (*net);
  if (minimal
      ? lnet.generateMinimal (thePrinter)
      : lnet.generate (thePrinter)) {
    unsigned* marking = lnet.unfold (*net->getInitMarking ());
    if (!marking) {
      thePrinter.printRaw ("assuming empty initial marking");
      thePrinter.finish ();
    }
    switch (format) {
    case fmtLoLA:
      lnet.toLoLA (f, marking);
      break;
    case fmtPEP:
      lnet.toPEP (f, marking);
      break;
    case fmtPROD:
      lnet.toPROD (f, f2, marking);
      break;
    case fmtDefault:
      thePrinter.setOutput (f);
      lnet.display (thePrinter, marking);
      thePrinter.setOutput (stderr);
    }
    delete[] marking;
  }
  else {
    thePrinter.printRaw ("unable to unfold the model");
    thePrinter.finish ();
  }
  if (f != stdout)
    fclose (f);
  if (f2 != stdout)
    fclose (f2);
}

void
dump (card_t visual)
{
  if (!net) {
    thePrinter.printRaw ("no model has been loaded");
    thePrinter.finish ();
  }
  else if (visual) {
    if (isState (0))
      vis (dump (visual > 1));
  }
  else {
    thePrinter.setBOL (0);
    net->display (thePrinter);
    thePrinter.setBOL (&demsg);
  }
}

/** Analyze the model using probabilistic methods
 * @param breadth	flag: search breadth first instead of depth
 */
static void
analyze (bool breadth)
{
  if (!net) {
    thePrinter.printRaw ("no model has been loaded");
    thePrinter.finish ();
    return;
  }

  class HashGraph hg;
  if (!hg.init (numTables, funcSize, hashSize)) {
    thePrinter.printRaw ("insufficient memory for ");
    thePrinter.print (card_t (numTables));
    thePrinter.printRaw (" tables of ");
    thePrinter.print (card_t (funcSize));
    thePrinter.printRaw (" functions and ");
    thePrinter.print (card_t (hashSize));
    thePrinter.printRaw (" bits");
    thePrinter.finish ();
    return;
  }

  if (verbose)
    fprintf (stderr, "%s:using probabilistic method -H%u,%u,%u\n", netname,
	     hashSize, funcSize, numTables);

#ifdef EXPR_COMPILE
  if (compilation)
    net->analyze (*compilation, thePrinter, hg, breadth);
  else
#endif // EXPR_COMPILE
    net->analyze (thePrinter, hg, breadth);

  thePrinter.setBOL (0);
  thePrinter.printQuoted (netname);
  thePrinter.printRaw (": at least ");
  thePrinter.print (hg.getNumStates ());
  thePrinter.printRaw (" states, ");
  thePrinter.print (hg.getNumArcs ());
  thePrinter.printRaw (" arcs");
  thePrinter.finish ();
  thePrinter.setBOL (&demsg);
}

void
analyze (card_t state, bool breadth)
{
  thePrinter.setBOL (0);
  if (!state && !graph) {
    analyze (breadth);
    thePrinter.setBOL (&demsg);
    return;
  }
  if (!isState (state)) {
    thePrinter.setBOL (&demsg);
    return;
  }
  if (graph->isProcessed (state)) {
    thePrinter.printRaw ("state ");
    thePrinter.printState (state);
    thePrinter.printRaw (" has already been analyzed");
    thePrinter.finish ();
  }
  if (verbose)
    fprintf (stderr, "%s:generating the reachability graph\n", netname);
#ifdef EXPR_COMPILE
  if (compilation)
    net->analyze (*compilation, thePrinter, *graph, state,
		  breadth ? Net::Breadth : Net::Depth);
  else
#endif // EXPR_COMPILE
    net->analyze (thePrinter, *graph, state,
		  breadth ? Net::Breadth : Net::Depth);

  stats ();
}

/** Open a file
 * @param name		name of the file
 * @param origfile	name of the original file
 * @param includes	include directories
 * @return		handle to the file, or NULL in case of error
 */
inline static FILE*
openfile (const char* name,
	  const char* origfile,
	  const class NameList& includes)
{
  FILE* f;
  // first try from the current directory
  if (!(f = fopen (name, "r"))) {
    // next, try from the directory of the original net

    size_t dirlength;
    char* fullname;
    size_t namelength = strlen (name);
    for (dirlength = 0; origfile[dirlength]; dirlength++);
    while (dirlength > 0 && origfile[dirlength] != '/')
      dirlength--;
    if (dirlength > 0) {
      fullname = new char[namelength + dirlength + 2];
      memcpy (fullname, origfile, dirlength + 1);
      memcpy (fullname + dirlength + 1, name, namelength + 1);
      f = fopen (fullname, "r");
      delete[] fullname;
      if (f)
	return f;
    }

    for (NameList::const_iterator i = includes.begin ();
	 i != includes.end (); i++) {
      const char* dirname = *i;
      size_t dirlength = strlen (dirname);
      char* fullname = new char[namelength + dirlength + 2];
      memcpy (fullname, dirname, dirlength);
      fullname[dirlength] = '/';
      memcpy (fullname + dirlength + 1, name, namelength + 1);
      f = fopen (fullname, "r");
      delete[] fullname;
      if (f)
	break;
    }
  }

  if (f) {
    register int ch = getc (f);
    if (ferror (f)) {
      fclose (f);
      f = 0;
    }
    else
      ungetc (ch, f);
  }

  return f;
}

FILE*
openfile (const char* name)
{
  return openfile (name, netname, includePath);
}

FILE*
openscript (const char* name)
{
  return openfile (name, scriptname, includePath);
}

bool
isState (card_t state)
{
  if (!graph) {
    thePrinter.printRaw ("no reachability graph");
    if (!net)
      thePrinter.printRaw (" (no model loaded)");
    thePrinter.finish ();
    return false;
  }
  if (state < graph->getNumStates ())
    return true;
  thePrinter.printRaw ("the graph only has ");
  thePrinter.print (card_t (graph->getNumStates ()));
  thePrinter.printRaw (" states");
  thePrinter.finish ();
  return false;
}

void
show (card_t state, card_t visual)
{
  if (!isState (state))
    return;
  if (visual) {
    vis (show (state, visual > 1));
    return;
  }
  thePrinter.setBOL (0);
  if (componentgraph) {
    card_t comp = componentgraph->getComponent (state);
    if (comp != CARD_T_MAX) {
      thePrinter.printComponent (comp);
      thePrinter.printRaw (":");
    }
  }
  card_t numSucc = 0;
  class GlobalMarking* m = graph->fetchState (state);
  card_t* succ = graph->getSuccessors (state);
  assert (!succ || *succ);
  if (succ) {
    numSucc = *succ;
    delete[] succ;
  }
  if (!numSucc)
    thePrinter.printRaw (graph->isProcessed (state)
			 ? "deadlock " : "unprocessed ");
  m->display (thePrinter);
  thePrinter.finish ();
  if (card_t numPred = graph->getNumPredecessors (state)) {
    thePrinter.print (numPred);
    thePrinter.printRaw (numPred == 1
			 ? " predecessor"
			 : " predecessors");
    thePrinter.finish ();
  }
  if (numSucc) {
    thePrinter.print (numSucc);
    thePrinter.printRaw (numSucc == 1
			 ? " successor"
			 : " successors");
    thePrinter.finish ();
  }
  delete m;
  thePrinter.setBOL (&demsg);
}

void
show (const card_t* states, card_t visual)
{
  assert (states && *states);
  if (*states == 1)
    show (states[1], visual);
  else if (visual)
    vis (show (states, visual > 1, true));
  else {
    thePrinter.setBOL (0);
    thePrinter.printRaw ("path  (");
    thePrinter.print (*states);
    thePrinter.printRaw (" nodes):");
    thePrinter++.linebreak ();
    printReversePath (states);
    thePrinter--.finish ();
    thePrinter.setBOL (&demsg);
  }
}

void
succ (card_t state, card_t visual)
{
  if (!isState (state))
    return;
  if (visual) {
    vis (succ (state, visual > 1));
    return;
  }
  thePrinter.setBOL (0);
#ifdef EXPR_COMPILE
  if (compilation)
    net->analyze (*compilation, thePrinter, *graph, state, Net::Single);
  else
#endif // EXPR_COMPILE
    net->analyze (thePrinter, *graph, state, Net::Single);
  word_t* data;
  card_t* succ = graph->getSuccessors (state, &data);
  assert (!succ || *succ);
  if (succ) {
    class BitUnpacker buf (data);
    for (card_t i = 1; i <= *succ; i++) {
      const class Transition* transition;
      class Valuation v;
      graph->getSuccessor (buf, transition, v);
      thePrinter.printRaw ("transition ");
      thePrinter.print (transition->getName ());
      thePrinter.printRaw ("->");
      thePrinter.printState (succ[i]);
      thePrinter.finish ();
      v.display (thePrinter);
      thePrinter.finish ();
    }
    delete[] data;
    delete[] succ;
  }
  thePrinter.setBOL (&demsg);
}

void
pred (card_t state, card_t visual)
{
  if (!isState (state))
    return;
  if (visual) {
    vis (pred (state, visual > 1));
    return;
  }
  thePrinter.setBOL (0);
  Graph::fpos_t pos = graph->getPredecessors (state);
  for (card_t oldsource = CARD_T_MAX; pos != FPOS_NONE; ) {
    card_t source = graph->getPredecessor (pos);
    if (oldsource == source)
      continue; /* several arcs from the same state */
    oldsource = source;
    word_t* data;
    card_t* succ = graph->getSuccessors (source, &data);
    assert (succ && *succ);
    if (succ) {
      class BitUnpacker buf (data);
      for (card_t i = 1; i <= *succ; i++) {
	const class Transition* transition;
	class Valuation v;
	graph->getSuccessor (buf, transition, v);
	if (succ[i] != state)
	  continue;
	thePrinter.printRaw ("transition ");
	thePrinter.print (transition->getName ());
	thePrinter.printRaw ("<-");
	thePrinter.printState (source);
	thePrinter.finish ();
	v.display (thePrinter);
	thePrinter.finish ();
      }
      delete[] data;
      delete[] succ;
    }
  }
  thePrinter.setBOL (&demsg);
}

void
stats (void)
{
  if (!isState (0))
    return;
  thePrinter.setBOL (0);
  thePrinter.printQuoted (graph->getFilename ());
  thePrinter.printRaw (": ");
  thePrinter.print (card_t (graph->getNumStates ()));
  thePrinter.printRaw (" states, ");
  thePrinter.print (graph->getNumArcs ());
  thePrinter.printRaw (" arcs");
  if (componentgraph) {
    thePrinter.printRaw (", ");
    thePrinter.print (componentgraph->size ());
    thePrinter.printRaw (componentgraph->size () == 1
			 ? " component"
			 : " components");
  }
  thePrinter.finish ();
  thePrinter.setBOL (&demsg);
}

void
showtimes (void)
{
#ifndef __unix
  clock_t nowclock = clock ();
  fprintf (stderr, "%lf seconds\n",
	   double (nowclock - lastclock) / CLOCKS_PER_SEC);
  lastclock = nowclock;
#else // !__unix
  static clock_t user, system;
  struct tms current;
  clock_t currentclock = times (&current);
  if (currentclock == -1) {
    perror ("times");
    return;
  }

  long ticks = sysconf (_SC_CLK_TCK);

  fprintf (stderr, "%lf real, %lf user, %lf system\n",
	   double (currentclock - lastclock) / ticks,
	   double (current.tms_utime - user) / ticks,
	   double (current.tms_stime - system) / ticks);
  user = current.tms_utime;
  system = current.tms_stime;
  lastclock = currentclock;
#endif // !__unix
}

void
strong (card_t state, class Expression* cond)
{
  if (!isState (state))
    return;
  if (componentgraph)
    delete componentgraph;
  componentgraph = new class ComponentGraph (*graph);
  unsigned comps = componentgraph->compute (state, cond);
  if (!comps && !interrupted) {
    thePrinter.printRaw ("failed to compute strongly connected components");
    thePrinter.finish ();
  }
  else if (comps) stats ();
  cond->destroy ();
}

/** Check if a component number exists
 * @param comp	the component
 * @return	true if the strongly connected component exists
 */
static bool
isComponent (card_t comp)
{
  if (!componentgraph) {
    thePrinter.printRaw ("strongly connected components "
			 "have not been computed");
    thePrinter.finish ();
    return false;
  }
  if (comp < componentgraph->size ())
    return true;
  thePrinter.printRaw ("there are only ");
  thePrinter.print (card_t (componentgraph->size ()));
  thePrinter.printRaw (" strongly connected components");
  thePrinter.finish ();
  return false;
}

void
terminal (void)
{
  if (!isComponent (0))
    return;
  thePrinter.setBOL (0);
  for (card_t comp = 0; comp < componentgraph->size (); comp++) {
    if (componentgraph->getNumSucc (comp))
      continue;
    card_t* states = componentgraph->getStates (comp);
    assert (states && *states && *states <= graph->getNumStates ());
    if (*states > 1 && *states != graph->getNumStates ()) {
      thePrinter.printRaw ("non-trivial terminal component ");
      thePrinter.printComponent (comp);
      thePrinter.printRaw (" (");
      thePrinter.print (*states);
      thePrinter.printRaw (" states):");
      thePrinter++.linebreak ();
      for (card_t state = 1;;) {
	thePrinter.printState (states[state]);
	if (state++ < *states)
	  thePrinter.delimiter (' ');
	else break;
      }
      thePrinter--.finish ();
    }
    delete[] states;
  }
  thePrinter.setBOL (&demsg);
}

void
allcomps (card_t visual)
{
  if (!isComponent (0))
    return;
  if (visual) {
    vis (components (visual > 1));
    return;
  }
  thePrinter.setBOL (0);
  for (card_t comp = 0; comp < componentgraph->size (); comp++) {
    card_t* states = componentgraph->getStates (comp);
    assert (states && *states && *states <= graph->getNumStates ());
    if (*states > 1 && *states != graph->getNumStates ())
      thePrinter.printRaw ("non-trivial ");
    if (!componentgraph->getNumSucc (comp))
      thePrinter.printRaw ("terminal ");
    thePrinter.printRaw ("component ");
    thePrinter.printComponent (comp);
    if (*states == graph->getNumStates ()) {
      thePrinter.printRaw (" (all states)");
      thePrinter.finish ();
      delete[] states;
      continue;
    }
    thePrinter.printRaw (" (");
    thePrinter.print (*states);
    thePrinter.printRaw (" states):");
    thePrinter++.linebreak ();
    for (card_t state = 1;;) {
      thePrinter.printState (states[state]);
      if (state++ < *states)
	thePrinter.delimiter (' ');
      else break;
    }
    thePrinter--.finish ();
    delete[] states;
  }
  thePrinter.setBOL (&demsg);
}

void
comp_show (card_t comp,
	   class Expression* cond,
	   card_t visual)
{
  if (!isComponent (comp)) {
    cond->destroy ();
    return;
  }
  if (visual) {
    vis (component (comp, cond, visual > 1));
    cond->destroy ();
    return;
  }
  thePrinter.setBOL (0);
  card_t* states = componentgraph->getStates (comp);
  assert (states && *states && *states <= graph->getNumStates ());
  if (!componentgraph->getNumSucc (comp))
    thePrinter.printRaw ("terminal ");
  if (!cond) {
    if (*states == 1) {
      thePrinter.printRaw ("trivial strongly connected component ");
      thePrinter.printComponent (comp);
      thePrinter.printRaw (": ");
      thePrinter.printState (states[1]);
      thePrinter.finish ();
      delete[] states;
      return;
    }
    else if (*states == graph->getNumStates ()) {
      thePrinter.printRaw ("trivial strongly connected component ");
      thePrinter.printComponent (comp);
      thePrinter.printRaw (": (all nodes)");
      thePrinter.finish ();
      delete[] states;
      return;
    }
  }
  thePrinter.printRaw ("strongly connected component ");
  thePrinter.printComponent (comp);
  thePrinter.printRaw (" (");
  thePrinter.print (*states);
  thePrinter.printRaw (" states):");
  thePrinter++.linebreak ();
  card_t state = 1;
  if (cond) {
    for (;;) {
      if (!graph->eval (states[state], cond)) {
	if (state++ < *states) continue; else break;
      }
      thePrinter.printState (states[state]);
      if (state++ < *states)
	thePrinter.delimiter (' ');
      else break;
    }
  }
  else {
    for (;;) {
      thePrinter.printState (states[state]);
      if (state++ < *states)
	thePrinter.delimiter (' ');
      else break;
    }
  }
  delete[] states;
  thePrinter--.finish ();
  thePrinter.setBOL (&demsg);
  cond->destroy ();
}

void
comp_succ (card_t comp, card_t visual)
{
  if (!isComponent (comp))
    return;
  if (visual) {
    vis (comp_succ (comp, visual > 1));
    return;
  }
  if (card_t* succ = componentgraph->getSucc (comp)) {
    assert (*succ > 0);
    thePrinter.setBOL (0);
    for (comp = 1;;) {
      thePrinter.printComponent (succ[comp]);
      if (comp++ < *succ)
	thePrinter.delimiter (' ');
      else break;
    }
    delete[] succ;
    thePrinter.finish ();
    thePrinter.setBOL (&demsg);
  }
}

void
comp_pred (card_t comp, card_t visual)
{
  if (!isComponent (comp))
    return;
  if (visual) {
    vis (comp_pred (comp, visual > 1));
    return;
  }
  if (card_t* pred = componentgraph->getPred (comp)) {
    assert (*pred > 0);
    thePrinter.setBOL (0);
    for (comp = 1;;) {
      thePrinter.printComponent (pred[comp]);
      if (comp++ < *pred)
	thePrinter.delimiter (' ');
      else break;
    }
    delete[] pred;
    thePrinter.finish ();
    thePrinter.setBOL (&demsg);
  }
}

static void
printLSTS (class LSTS& lsts,
	   const card_t* path,
	   bool reverse)
{
  class BitVector visited (graph->getNumStates ());
  for (const card_t* p = path + *path; p > path; p--) {
    if (!visited.tset (*p)) {
      class GlobalMarking* m = graph->fetchState (*p);
      lsts.outputState (*m, *p);
      delete m;
    }
    if (const card_t* next = reverse
	? (p < path + *path ? p + 1 : 0)
	: (p > path + 1 ? p - 1 : 0)) {
      word_t* data;
      card_t* succ = graph->getSuccessors (*p, &data);
      assert (!succ || *succ);
      if (succ) {
	class BitUnpacker buf (data);
	for (card_t i = 1; i <= *succ; i++) {
	  const class Transition* transition;
	  class Valuation v;
	  graph->getSuccessor (buf, transition, v);
	  if (succ[i] != *next)
	    continue;
	  lsts.outputArc (*p, *next, *transition, v);
	}
	delete[] data;
	delete[] succ;
      }
    }
  }
}

static void
printPath (const card_t* path)
{
  assert (path && graph);
  if (class LSTS* lsts = graph->getLSTS ()) {
    printLSTS (*lsts, path, false);
    graph->setLSTS (0);
  }
  for (const card_t* p = path + *path;;) {
    thePrinter.printState (*p);
    if (--p > path)
      thePrinter.delimiter (' ');
    else break;
  }
}

static void
printReversePath (const card_t* path)
{
  assert (path && graph);
  if (class LSTS* lsts = graph->getLSTS ()) {
    printLSTS (*lsts, path, true);
    graph->setLSTS (0);
  }
  for (card_t i = 1;;) {
    thePrinter.printState (path[i]);
    if (i++ < *path)
      thePrinter.delimiter (' ');
    else break;
  }
}

void
path_sc (card_t state, card_t comp,
	 class Expression* pathc,
	 card_t visual)
{
  if (!isState (state) || !isComponent (comp)) {
    pathc->destroy ();
    return;
  }
  thePrinter.setBOL (0);
  if (card_t* path = componentgraph->path (state, comp, pathc)) {
    if (visual)
      vis (show (path, visual > 1, false));
    else {
      thePrinter.printRaw ("shortest path from ");
      thePrinter.printState (state);
      thePrinter.printRaw (" to ");
      thePrinter.printComponent (comp);
      thePrinter.printRaw (" (");
      thePrinter.print (*path);
      thePrinter.printRaw (" nodes):");
      thePrinter++.linebreak ();
      printPath (path);
      thePrinter--.finish ();
    }
    delete[] path;
  }
  else {
    thePrinter.printRaw (pathc ? "no such path from " : "no path from ");
    thePrinter.printState (state);
    thePrinter.printRaw (" to ");
    thePrinter.printComponent (comp);
    thePrinter.finish ();
  }
  thePrinter.setBOL (&demsg);
  pathc->destroy ();
}

void
path_cs (card_t comp, card_t state,
	 class Expression* pathc,
	 card_t visual)
{
  if (!isState (state) || !isComponent (comp)) {
    pathc->destroy ();
    return;
  }
  thePrinter.setBOL (0);
  if (card_t* path = componentgraph->rpath (state, comp, pathc)) {
    if (visual)
      vis (show (path, visual > 1, true));
    else {
      thePrinter.printRaw ("shortest path from ");
      thePrinter.printComponent (comp);
      thePrinter.printRaw (" to ");
      thePrinter.printState (state);
      thePrinter.printRaw (" (");
      thePrinter.print (*path);
      thePrinter.printRaw (" nodes):");
      thePrinter++.linebreak ();
      printReversePath (path);
      thePrinter--.finish ();
    }
    delete[] path;
  }
  else {
    thePrinter.printRaw (pathc ? "no such path from " : "no path from ");
    thePrinter.printComponent (comp);
    thePrinter.printRaw (" to ");
    thePrinter.printState (state);
    thePrinter.finish ();
  }
  thePrinter.setBOL (&demsg);
  pathc->destroy ();
}

void
path_ss (card_t state, card_t target,
	 class Expression* pathc,
	 card_t visual)
{
  if (!isState (state) || !isState (target)) {
    pathc->destroy ();
    return;
  }
  thePrinter.setBOL (0);
  if (card_t* path = graph->path (state, target, pathc)) {
    if (visual)
      vis (show (path, visual > 1, false));
    else {
      thePrinter.printRaw ("shortest path from ");
      thePrinter.printState (state);
      thePrinter.printRaw (" to ");
      thePrinter.printState (target);
      thePrinter.printRaw (" (");
      thePrinter.print (*path);
      thePrinter.printRaw (" nodes):");
      thePrinter++.linebreak ();
      printPath (path);
      thePrinter--.finish ();
    }
    delete[] path;
  }
  else {
    thePrinter.printRaw (pathc ? "no such path from " : "no path from ");
    thePrinter.printState (state);
    thePrinter.printRaw (" to ");
    thePrinter.printState (target);
    thePrinter.finish ();
  }
  thePrinter.setBOL (&demsg);
  pathc->destroy ();
}

void
path_se (card_t state, class Expression& cond,
	 class Expression* pathc,
	 card_t visual)
{
  if (!isState (state)) {
    cond.destroy ();
    pathc->destroy ();
    return;
  }
  thePrinter.setBOL (0);
  if (card_t* path = graph->path (state, cond, pathc)) {
    if (visual)
      vis (show (path, visual > 1, false));
    else {
      thePrinter.printRaw ("shortest path from ");
      thePrinter.printState (state);
      thePrinter.printRaw (" to condition (");
      thePrinter.print (*path);
      thePrinter.printRaw (" nodes):");
      thePrinter++.linebreak ();
      printPath (path);
      thePrinter--.finish ();
    }
    delete[] path;
  }
  else {
    thePrinter.printRaw (pathc ? "no such path from " : "no path from ");
    thePrinter.printState (state);
    thePrinter.printRaw (" to condition");
    thePrinter.finish ();
  }
  thePrinter.setBOL (&demsg);
  cond.destroy ();
  pathc->destroy ();
}

void
rpath_se (card_t state, class Expression& cond,
	  class Expression* pathc,
	  card_t visual)
{
  if (!isState (state)) {
    cond.destroy ();
    pathc->destroy ();
    return;
  }
  thePrinter.setBOL (0);
  if (card_t* path = graph->rpath (state, cond, pathc)) {
    if (visual)
      vis (show (path, visual > 1, true));
    else {
      thePrinter.printRaw ("shortest path from condition to ");
      thePrinter.printState (state);
      thePrinter.printRaw (" (");
      thePrinter.print (*path);
      thePrinter.printRaw (" nodes):");
      thePrinter++.linebreak ();
      printReversePath (path);
      thePrinter--.finish ();
    }
    delete[] path;
  }
  else {
    thePrinter.printRaw (pathc
			 ? "no such path from condition to "
			 : "no path from condition to ");
    thePrinter.printState (state);
    thePrinter.finish ();
  }
  thePrinter.setBOL (&demsg);
  cond.destroy ();
  pathc->destroy ();
}

void
path_sl (card_t state,
	 const card_t* loop,
	 class Expression* pathc,
	 card_t visual)
{
  if (!isState (state)) {
    pathc->destroy ();
    return;
  }
  thePrinter.setBOL (0);
  if (card_t* path = graph->path (state, loop, pathc)) {
    if (visual)
      vis (show (path, visual > 1, false));
    else {
      thePrinter.printRaw ("shortest path from ");
      thePrinter.printState (state);
      thePrinter.printRaw (" to loop (");
      thePrinter.print (*path);
      thePrinter.printRaw (" nodes):");
      thePrinter++.linebreak ();
      printPath (path);
      thePrinter--.finish ();
    }
    delete[] path;
  }
  else {
    thePrinter.printRaw (pathc
			 ? "no such path from "
			 : "no path from ");
    thePrinter.printState (state);
    thePrinter.printRaw (" to loop");
    thePrinter.finish ();
  }
  thePrinter.setBOL (&demsg);
  pathc->destroy ();
}

void
eval (card_t state, class Expression* expr, card_t visual)
{
  if (!isState (state))
    return;
  thePrinter.setBOL (0);
  class Valuation v;
  if (!expr);
  else if (expr->isSet ()) {
    class GlobalMarking* m = graph->fetchState (state);
    v.setGlobalMarking (m);
    if (class PlaceMarking* p = expr->meval (v)) {
      p->display (thePrinter);
      delete p;
    }
    else
      v.display (thePrinter);
    thePrinter.finish ();
    delete m;
  }
  else if (!expr->isTemporal ()) {
    class GlobalMarking* m = graph->fetchState (state);
    v.setGlobalMarking (m);
    if (class Value* value = expr->eval (v)) {
      value->display (thePrinter);
      delete value;
    }
    else
      v.display (thePrinter);
    thePrinter.finish ();
    delete m;
  }
  else {
    thePrinter.setBOL (&demsg);
    onthefly (state, expr, visual);
  }
  expr->destroy ();
  thePrinter.setBOL (&demsg);
}
